# sage.doctest: needs sage.libs.pari
r"""
Interface to the GP calculator of PARI/GP

Type ``gp.[tab]`` for a list of all the functions
available from your Gp install. Type ``gp.[tab]?`` for
Gp's help about a given function. Type ``gp(...)`` to
create a new Gp object, and ``gp.eval(...)`` to evaluate a
string using Gp (and get the result back as a string).

EXAMPLES: We illustrate objects that wrap GP objects (gp is the
PARI interpreter)::

    sage: M = gp('[1,2;3,4]')
    sage: M
    [1, 2; 3, 4]
    sage: M * M
    [7, 10; 15, 22]
    sage: M + M
    [2, 4; 6, 8]
    sage: M.matdet()
    -2

::

    sage: E = gp.ellinit([1,2,3,4,5])
    sage: E.ellglobalred()
    [10351, [1, -1, 0, -1], 1, [11, 1; 941, 1], [[1, 5, 0, 1], [1, 5, 0, 1]]]
    sage: E.ellan(20)
    [1, 1, 0, -1, -3, 0, -1, -3, -3, -3, -1, 0, 1, -1, 0, -1, 5, -3, 4, 3]

::

    sage: primitive_root(7)
    3
    sage: x = gp("znlog( Mod(2,7), Mod(3,7))")
    sage: 3^x % 7
    2

::

    sage: print(gp("taylor(sin(x),x)"))
    x - 1/6*x^3 + 1/120*x^5 - 1/5040*x^7 + 1/362880*x^9 - 1/39916800*x^11 + 1/6227020800*x^13 - 1/1307674368000*x^15 + O(x^16)

GP has a powerful very efficient algorithm for numerical
computation of integrals.

::

    sage: gp("a = intnum(x=0,6,sin(x))")
    0.039829713349633979454347702077075594548
    sage: gp("a")
    0.039829713349633979454347702077075594548
    sage: gp.kill("a")
    sage: gp("a")
    a

Note that gp ASCII plots *do* work in Sage, as follows::

    sage: print(gp.eval("plot(x=0,6,sin(x))"))
    <BLANKLINE>
    0.9988963 |''''''''''''_x...x_''''''''''''''''''''''''''''''''''''''''''|
              |          x"        "x                                        |
              |        _"            "_                                      |
              |       x                x                                     |
              |      "                  "                                    |
              |     "                    "                                   |
              |   _"                      "_                                 |
              |  _                          _                                |
              | _                            _                               |
              |_                              _                              |
              _                                                              |
              `````````````````````````````````"``````````````````````````````
              |                                 "                            |
              |                                  "                           |
              |                                   "                          "
              |                                    "_                      _"|
              |                                      _                    _  |
              |                                       _                  _   |
              |                                        x                x    |
              |                                         "_            _"     |
              |                                           x_        _x       |
    -0.998955 |............................................."x____x".........|
              0                                                              6

The GP interface reads in even very long input (using files) in a
robust manner, as long as you are creating a new object.

::

    sage: t = '"%s"'%10^10000   # ten thousand character string.
    sage: a = gp.eval(t)
    sage: a = gp(t)

In Sage, the PARI large Galois groups datafiles should be installed
by default::

    sage: f = gp('x^9 - x - 2')
    sage: f.polgalois()
    [362880, -1, 34, "S9"]

TESTS:

Test error recovery::

    sage: x = gp('1/0')
    Traceback (most recent call last):
    ...
    TypeError: Error executing code in GP:
    CODE:
        sage[...]=1/0;
    PARI/GP ERROR:
      ***   at top-level: sage[...]=1/0
      ***                          ^--
      *** _/_: impossible inverse in gdiv: 0.

AUTHORS:

- William Stein

- David Joyner: some examples

- William Stein (2006-03-01): added tab completion for methods:
  gp.[tab] and x = gp(blah); x.[tab]

- William Stein (2006-03-01): updated to work with PARI 2.2.12-beta

- William Stein (2006-05-17): updated to work with PARI 2.2.13-beta
"""

##########################################################################
#
#       Copyright (C) 2005 William Stein <wstein@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#                  https://www.gnu.org/licenses/
#
##########################################################################
import os

import sage.interfaces.abc
from sage.env import DOT_SAGE
from sage.interfaces.tab_completion import ExtraTabCompletion
from sage.libs.pari import pari
from sage.misc.instancedoc import instancedoc
from sage.misc.lazy_import import lazy_import
from sage.misc.verbose import verbose

lazy_import('sage.rings.cc', 'CC')

from sage.interfaces.expect import (
    Expect,
    ExpectElement,
    ExpectFunction,
    FunctionElement,
)


class Gp(ExtraTabCompletion, Expect):
    """
    Interface to the PARI gp interpreter.

    Type ``gp.[tab]`` for a list of all the functions
    available from your Gp install. Type ``gp.[tab]?`` for
    Gp's help about a given function. Type ``gp(...)`` to
    create a new Gp object, and ``gp.eval(...)`` to evaluate a
    string using Gp (and get the result back as a string).

        INPUT:

        - ``stacksize`` -- integer (default: 10000000); the initial PARI
          stacksize in bytes (default: 10MB)
        - ``script_subdirectory`` -- string (default: ``None``); name of the
          subdirectory of ``SAGE_EXTCODE/pari`` from which to read scripts
        - ``logfile`` -- string (default: ``None``); log file for the pexpect
          interface
        - ``server`` -- name of remote server
        - ``server_tmpdir`` -- name of temporary directory on remote server
        - ``init_list_length`` -- integer (default: 1024); length of initial
          list of local variables
        - ``seed`` -- integer (default: random); random number generator seed
          for pari

        EXAMPLES::

            sage: Gp()
            PARI/GP interpreter
    """
    def __init__(self, stacksize=10000000,   # 10MB
                 maxread=None, script_subdirectory=None,
                 logfile=None,
                 server=None,
                 server_tmpdir=None,
                 init_list_length=1024,
                 seed=None):
        """
        Initialization of this PARI gp interpreter.

        INPUT:

        - ``stacksize`` -- integer (default: 10000000); the initial PARI
          stacksize in bytes (default: 10MB)
        - ``script_subdirectory`` -- string (default: ``None``); name of the
          subdirectory of SAGE_EXTCODE/pari from which to read scripts
        - ``logfile`` -- string (default: ``None``); log file for the pexpect
          interface
        - ``server`` -- name of remote server
        - ``server_tmpdir`` -- name of temporary directory on remote server
        - ``init_list_length`` -- integer (default: 1024); length of initial
          list of local variables.
        - ``seed`` -- integer (default random nonzero 31 bit integer); value of
          random seed

        EXAMPLES::

            sage: from sage.interfaces.gp import gp
            sage: gp == loads(dumps(gp))
            True
        """
        Expect.__init__(self,
                        name='pari',
                        prompt='\\? ',
                        # --fast so the system gprc isn't read (we configure below)
                        command=f"gp --fast --emacs --quiet --stacksize {stacksize}",
                        maxread=maxread,
                        server=server,
                        server_tmpdir=server_tmpdir,
                        script_subdirectory=script_subdirectory,
                        restart_on_ctrlc=False,
                        verbose_start=False,
                        logfile=logfile,
                        eval_using_file_cutoff=1024)
        self.__seq = 0
        self.__var_store_len = 0
        self.__init_list_length = init_list_length
        self._seed = seed

    def set_seed(self, seed=None):
        """
        Set the seed for gp interpreter.

        The seed should be an integer.

        EXAMPLES::

            sage: g = Gp()
            sage: g.set_seed(1)
            1
            sage: [g.random() for i in range(5)]
            [1546275796, 879788114, 1745191708, 771966234, 1247963869]
        """
        if seed is None:
            seed = self.rand_seed()
        self.eval("setrand(%d)" % seed)
        self._seed = seed
        return seed

    def _start(self, alt_message=None, block_during_init=True):
        Expect._start(self, alt_message, block_during_init)
        # disable memory debugging: those warnings can only confuse our
        # interface
        self._eval_line('default(debugmem,0);')
        # disable timer
        self._eval_line('default(timer,0);')
        # disable the break loop, otherwise gp will seem to hang on errors
        self._eval_line('default(breakloop,0);')
        # list of directories where gp will look for scripts (only current working directory)
        self._eval_line('default(path,".");')
        # executable for gp ?? help
        self._eval_line('default(help, "gphelp -detex");')
        # logfile disabled since Expect already logs
        self._eval_line('default(log,0);')
        self._eval_line("default(nbthreads,1);")
        # set random seed
        self.set_seed(self._seed)

    def _repr_(self):
        """
        String representation of this PARI gp interpreter.

        EXAMPLES::

            sage: gp  # indirect doctest
            PARI/GP interpreter
        """
        return 'PARI/GP interpreter'

    def __reduce__(self):
        """
        EXAMPLES::

            sage: from sage.interfaces.gp import gp
            sage: gp.__reduce__()
            (<function reduce_load_GP at 0x...>, ())
            sage: f, args = _
            sage: f(*args)
            PARI/GP interpreter
        """
        return reduce_load_GP, tuple([])

    def _function_class(self):
        """
        Return the GpFunction class.

        EXAMPLES::

            sage: gp._function_class()
            <class 'sage.interfaces.expect.ExpectFunction'>
            sage: type(gp.gcd)
            <class 'sage.interfaces.expect.ExpectFunction'>
        """
        return GpFunction

    def _quit_string(self):
        r"""
        Return the string used to quit the GP interpreter.

        EXAMPLES::

            sage: gp._quit_string()
            '\\q'

        ::

            sage: g = Gp()
            sage: a = g(2)
            sage: g.is_running()
            True
            sage: g.quit()
            sage: g.is_running()
            False
        """
        return r"\q"

    def _read_in_file_command(self, filename):
        r"""
        Return the string used to read filename into GP.

        EXAMPLES::

            sage: gp._read_in_file_command('test')
            'read("test")'

        ::

            sage: filename = tmp_filename()
            sage: f = open(filename, 'w')
            sage: _ = f.write('x = 22;\n')
            sage: f.close()
            sage: gp.read(filename)
            sage: gp.get('x').strip()
            '22'
        """
        return 'read("%s")' % filename

    def _tab_completion(self):
        """
        EXAMPLES::

            sage: c = gp._tab_completion()
            sage: len(c) > 100
            True
            sage: 'gcd' in c
            True
        """
        try:
            b = self.__builtin
        except AttributeError:
            b = self.eval('?*').split()
            self.__builtin = b
        return b + self.eval('?0').split()

    def get_precision(self):
        """
        Return the current PARI precision for real number computations.

        EXAMPLES::

            sage: gp.get_precision()
            38
        """
        return self.get_default('realprecision')

    get_real_precision = get_precision

    def set_precision(self, prec):
        """
        Set the PARI precision (in decimal digits) for real
        computations, and returns the old value.

        .. NOTE::

            PARI/GP rounds up precisions to the nearest machine word,
            so the result of :meth:`get_precision` is not always the
            same as the last value inputted to :meth:`set_precision`.

        EXAMPLES::

            sage: old_prec = gp.set_precision(53); old_prec
            38
            sage: gp.get_precision()
            57
            sage: gp.set_precision(old_prec)
            57
            sage: gp.get_precision()
            38
        """
        return self.set_default('realprecision', prec)

    set_real_precision = set_precision

    def get_series_precision(self):
        """
        Return the current PARI power series precision.

        EXAMPLES::

            sage: gp.get_series_precision()
            16
        """
        return self.get_default('seriesprecision')

    def set_series_precision(self, prec=None):
        """
        Set the PARI power series precision, and returns the old precision.

        EXAMPLES::

            sage: old_prec = gp.set_series_precision(50); old_prec
            16
            sage: gp.get_series_precision()
            50
            sage: gp.set_series_precision(old_prec)
            50
            sage: gp.get_series_precision()
            16
        """
        return self.set_default('seriesprecision', prec)

    def _eval_line(self, line, allow_use_file=True, wait_for_prompt=True, restart_if_needed=False):
        """
        EXAMPLES::

            sage: gp._eval_line('2+2')
            '4'

        TESTS:

        We verify that :issue:`11617` is fixed::

            sage: gp._eval_line('a='+str(list(range(2*10^5))))[:70]
            '[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,'
        """
        line = line.strip()
        if len(line) == 0:
            return ''
        a = Expect._eval_line(self, line,
                              allow_use_file=allow_use_file,
                              wait_for_prompt=wait_for_prompt)
        if a.find("the PARI stack overflows") != -1:
            verbose("automatically doubling the PARI stack and re-executing current input line")
            b = self.eval("allocatemem()")
            if b.find("Warning: not enough memory") != -1:
                raise RuntimeError(a)
            return self._eval_line(line, allow_use_file=allow_use_file,
                                   wait_for_prompt=wait_for_prompt)
        return a

    def cputime(self, t=None):
        """
        cputime for pari - cputime since the pari process was started.

        INPUT:

        - ``t`` -- (default: ``None``) if not None, then returns
          time since t


        .. warning::

           If you call gettime explicitly, e.g., gp.eval('gettime'),
           you will throw off this clock.

        EXAMPLES::

            sage: gp.cputime()          # random output
            0.0080000000000000002
            sage: gp.factor('2^157-1')
            [852133201, 1; 60726444167, 1; 1654058017289, 1; 2134387368610417, 1]
            sage: gp.cputime()          # random output
            0.26900000000000002
        """
        try:
            tm = self._last
        except AttributeError:
            tm = 0.0
        m = eval(self.eval('gettime()/1000.0')) + tm
        self._last = m
        if t:
            return m - t
        return m

    def set_default(self, var, value):
        """
        Set a PARI gp configuration variable, and return the old value.

        INPUT:

        - ``var`` -- string; the name of a PARI gp
          configuration variable (see ``gp.default()`` for a list)
        - ``value`` -- the value to set the variable to

        EXAMPLES::

            sage: old_prec = gp.set_default('realprecision', 110)
            sage: gp.get_default('realprecision')
            115
            sage: gp.set_default('realprecision', old_prec)
            115
            sage: gp.get_default('realprecision')
            38
        """
        old = self.get_default(var)
        self._eval_line('default(%s,%s)' % (var, value))
        return old

    def get_default(self, var):
        """
        Return the current value of a PARI gp configuration variable.

        INPUT:

        - ``var`` -- string; the name of a PARI gp
          configuration variable (see ``gp.default()`` for a list)

        OUTPUT: string; the value of the variable

        EXAMPLES::

            sage: gp.get_default('log')
            0
            sage: gp.get_default('datadir')
            '.../share/pari'
            sage: gp.get_default('seriesprecision')
            16
            sage: gp.get_default('realprecision')
            38
        """
        return eval(self._eval_line('default(%s)' % var))

    def set(self, var, value):
        """
        Set the GP variable var to the given value.

        INPUT:

        - ``var`` -- string; a valid GP variable identifier
        - ``value`` -- a value for the variable

        EXAMPLES::

            sage: gp.set('x', '2')
            sage: gp.get('x')
            '2'
        """
        cmd = '%s=%s;' % (var, value)
        out = self.eval(cmd)
        if out.find('***') != -1:
            raise TypeError("Error executing code in GP:\nCODE:\n\t%s\nPARI/GP ERROR:\n%s" % (cmd, out))

    def get(self, var):
        """
        Get the value of the GP variable var.

        INPUT:

        - ``var`` -- string; a valid GP variable identifier

        EXAMPLES::

            sage: gp.set('x', '2')
            sage: gp.get('x')
            '2'
        """
        return self.eval('print(%s)' % var)

    def kill(self, var):
        """
        Kill the value of the GP variable var.

        INPUT:

        - ``var`` -- string; a valid GP variable identifier

        EXAMPLES::

            sage: gp.set('xx', '22')
            sage: gp.get('xx')
            '22'
            sage: gp.kill('xx')
            sage: gp.get('xx')
            'xx'
        """
        self.eval('kill(%s)' % var)

    def _next_var_name(self):
        """
        Return the name of the next unused interface variable name.

        EXAMPLES::

            sage: g = Gp()
            sage: g._next_var_name()
            'sage[1]'
            sage: g(2)^2
            4
            sage: g._next_var_name()
            'sage[5]'

        TESTS:

        The vector of results is correctly resized when the stack has
        to be enlarged during this operation::

            sage: g = Gp(stacksize=3*10^6,init_list_length=12000)  # long time
            sage: for n in [1..13000]:  # long time
            ....:     a = g(n)
            sage: g('length(sage)')     # long time
            24000
        """
        self.__seq += 1
        if self.__seq >= self.__var_store_len:
            if self.__var_store_len == 0:
                self.eval('sage=vector(%s,k,0);' % self.__init_list_length)
                self.__var_store_len = self.__init_list_length
            else:
                self.eval('sage0=concat(sage, vector(%s,k,0));' % self.__var_store_len)
                self.eval('sage=sage0;')
                self.eval('kill(sage0);')
                self.__var_store_len *= 2
                verbose("doubling PARI/sage object vector: %s" % self.__var_store_len)
        return 'sage[%s]' % self.__seq

    def _reset_expect(self):
        """
        Reset state of the GP interface.

        EXAMPLES::

            sage: a = gp('10'); a
            10
            sage: gp.quit()  # indirect doctest
            sage: a
            (invalid PARI/GP interpreter object -- The pari session in which this object was defined is no longer running.)
            sage: gp("30!")
            265252859812191058636308480000000
        """
        self.__var_store_len = 0
        Expect._reset_expect(self)

    def console(self):
        """
        Spawn a new GP command-line session.

        EXAMPLES::

            sage: gp.console()  # not tested
            GP/PARI CALCULATOR Version 2.4.3 (development svn-12577)
            amd64 running linux (x86-64/GMP-4.2.1 kernel) 64-bit version
            compiled: Jul 21 2010, gcc-4.6.0 20100705 (experimental) (GCC)
            (readline v6.0 enabled, extended help enabled)
        """
        gp_console()

    def version(self):
        """
        Return the version of GP being used.

        EXAMPLES::

            sage: gp.version()  # not tested
            ((2, 4, 3), 'GP/PARI CALCULATOR Version 2.4.3 (development svn-12577)')
        """
        return gp_version()

    def _object_class(self):
        """
        Return the GpElement class.

        EXAMPLES::

            sage: gp._object_class()
            <class 'sage.interfaces.gp.GpElement'>
            sage: type(gp(2))
            <class 'sage.interfaces.gp.GpElement'>
        """
        return GpElement

    def _function_element_class(self):
        """
        Return the GpFunctionElement class.

        EXAMPLES::

            sage: gp._function_element_class()
            <class 'sage.interfaces.expect.FunctionElement'>
            sage: type(gp(2).gcd)
            <class 'sage.interfaces.expect.FunctionElement'>
        """
        return GpFunctionElement

    def _true_symbol(self):
        """
        Return the symbol used for truth in GP.

        EXAMPLES::

            sage: gp._true_symbol()
            '1'

        ::

            sage: gp(2) == gp(2)
            True
        """
        return '1'

    def _false_symbol(self):
        """
        Return the symbol used for falsity in GP.

        EXAMPLES::

            sage: gp._false_symbol()
            '0'

        ::

            sage: gp(2) == gp(3)
            False
        """
        return '0'

    def _equality_symbol(self):
        """
        Return the symbol used for equality in GP.

        EXAMPLES::

            sage: gp._equality_symbol()
            '=='

        ::

            sage: gp(2) == gp(2)
            True
        """
        return '=='

    def _exponent_symbol(self):
        """
        Return the symbol to denote the exponent of a number in GP.

        EXAMPLES::

            sage: gp._exponent_symbol()
            ' E'

        ::

            sage: repr(gp(10.^80)).replace(gp._exponent_symbol(), 'e')
            '1.0000000000000000000000000000000000000e80'
        """
        return ' E'

    def help(self, command):
        r"""
        Return GP's help for ``command``.

        EXAMPLES::

            sage: gp.help('gcd')
            'gcd(x,{y}): greatest common divisor of x and y.'
        """
        return self.eval('?%s' % command).strip()

    def new_with_bits_prec(self, s, precision=0):
        r"""
        Create a GP object from s with ``precision`` bits of
        precision. GP actually automatically increases this precision to
        the nearest word (i.e. the next multiple of 32 on a 32-bit machine,
        or the next multiple of 64 on a 64-bit machine).

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: pi_def = gp(pi); pi_def
            3.1415926535897932384626433832795028842
            sage: pi_def.precision()
            38
            sage: pi_150 = gp.new_with_bits_prec(pi, 150)
            sage: new_prec = pi_150.precision();
            sage: prec32 = 48
            sage: prec64 = 57
            sage: new_prec in [prec32, prec64]  # 32 vs 64 bit answers
            True
            sage: old_prec = gp.set_precision(new_prec); old_prec
            38
            sage: pi_150
            3.1415926535897932384626433832795028841971693993...
            sage: gp.set_precision(old_prec) in [prec32, prec64]
            True
            sage: gp.get_precision()
            38
        """
        if precision:
            old_prec = self.get_real_precision()
            prec = int(precision / 3.321928095)
            self.set_real_precision(prec)
            x = self(s)
            self.set_real_precision(old_prec)
        else:
            x = self(s)
        return x


@instancedoc
class GpElement(ExpectElement, sage.interfaces.abc.GpElement):
    """
    EXAMPLES: This example illustrates dumping and loading GP elements
    to compressed strings.

    ::

        sage: a = gp(39393)
        sage: loads(a.dumps()) == a
        True

    Since dumping and loading uses the string representation of the
    object, it need not result in an identical object from the point of
    view of PARI::

        sage: E = gp('ellinit([1,2,3,4,5])')
        sage: loads(dumps(E)) == E
        True
        sage: x = gp.Pi()/3
        sage: loads(dumps(x)) == x
        False
        sage: x
        1.0471975511965977461542144610931676281
        sage: loads(dumps(x))
        1.0471975511965977461542144610931676281

    The two elliptic curves look the same, but internally the floating
    point numbers are slightly different.
    """
    def _reduce(self):
        """
        Return the string representation of self, for pickling.

        Because the internal representation of a gp element is richer
        than the corresponding sage object, we use the string representation
        for pickling.

        EXAMPLES::

            sage: E = gp('ellinit([1,2,3,4,5])')
            sage: loads(dumps(E)) == E # indirect doctest
            True
            sage: gp(E.sage()) == E
            False
        """
        return repr(self)

    def _sage_(self):
        """
        Convert this GpElement into a Sage object, if possible.

        EXAMPLES::

            sage: gp(SR(I)).sage()                                                      # needs sage.symbolic
            i
            sage: gp(SR(I)).sage().parent()                                             # needs sage.symbolic
            Number Field in i with defining polynomial x^2 + 1 with i = 1*I

        ::

            sage: # needs sage.modules
            sage: M = Matrix(ZZ,2,2,[1,2,3,4]); M
            [1 2]
            [3 4]
            sage: gp(M)
            [1, 2; 3, 4]
            sage: gp(M).sage()
            [1 2]
            [3 4]
            sage: gp(M).sage() == M
            True

        Conversion of strings::

           sage: s = gp('"foo"')
           sage: s.sage()
           'foo'
           sage: type(s.sage())
           <class 'str'>
        """
        if self.is_string():
            return str(self)
        return pari(str(self)).sage()

    def is_string(self):
        """
        Tell whether this element is a string.

        EXAMPLES::

            sage: gp('"abc"').is_string()
            True
            sage: gp('[1,2,3]').is_string()
            False
        """
        return repr(self.type()) == 't_STR'

    def __float__(self):
        """
        Return Python float.

        EXAMPLES::

            sage: float(gp(10))
            10.0
        """
        return float(pari(str(self)))

    def __bool__(self):
        """
        EXAMPLES::

            sage: gp(2).bool()
            True
            sage: bool(gp(2))
            True
            sage: bool(gp(0))
            False
        """
        P = self._check_valid()
        return P.eval('%s != 0' % (self.name())) == '1'

    def _complex_mpfr_field_(self, CC):
        """
        Return ComplexField element of ``self``.

        INPUT:

        - ``CC`` -- a Complex or Real Field

        EXAMPLES::

            sage: # needs sage.symbolic
            sage: z = gp(SR(1+15*I)); z
            1 + 15*I
            sage: z._complex_mpfr_field_(CC)
            1.00000000000000 + 15.0000000000000*I
            sage: CC(z) # CC(gp(1+15*I))
            1.00000000000000 + 15.0000000000000*I
            sage: CC(gp(11243.9812+15*I))
            11243.9812000000 + 15.0000000000000*I
            sage: ComplexField(10)(gp(11243.9812+15*I))
            11000. + 15.*I
        """
        # Multiplying by CC(1) is necessary here since
        # sage: pari(gp(1+I)).sage().parent()
        # Maximal Order in Number Field in i with defining polynomial x^2 + 1
        return CC((CC.one() * pari(self)).sage())

    def _complex_double_(self, CDF):
        """
        Return this value as a CDF element.

        EXAMPLES::

            sage: CDF(gp(pi+I*e))                                                       # needs sage.symbolic
            3.141592653589793 + 2.718281828459045*I
        """
        # Retrieving values from another computer algebra system is
        # slow anyway, right?
        cc_val = self._complex_mpfr_field_(CC)
        return CDF(cc_val)

    def __len__(self):
        """
        EXAMPLES::

            sage: len(gp([1,2,3]))
            3
        """
        return int(self.length())

    def __del__(self):
        """
        Note that clearing object is pointless since it wastes time and
        PARI/GP doesn't really free used memory.

        EXAMPLES::

            sage: a = gp(2)
            sage: a.__del__()
            sage: a
            2
            sage: del a
            sage: a
            Traceback (most recent call last):
            ...
            NameError: name 'a' is not defined
        """
        return

    # This is tempting -- but the (la)tex output is very very
    # out of date, e.g., for matrices it uses \pmatrix (which
    # causes an error if amsmath is loaded) and for rationals
    # it does nothing, etc.
    # def _latex_(self):
    #     P = self._check_valid()
    #     return P.eval('printtex(%s)'%self.name())

    def _tab_completion(self):
        """
        EXAMPLES::

            sage: 'gcd' in gp(2)._tab_completion()
            True
        """
        return self.parent()._tab_completion()


GpFunctionElement = FunctionElement
GpFunction = ExpectFunction


# An instance
gp = Gp(logfile=os.path.join(DOT_SAGE, 'gp-expect.log'))
# useful for debugging!


def reduce_load_GP():
    """
    Return the GP interface object defined in ``sage.interfaces.gp``.

    EXAMPLES::

        sage: from sage.interfaces.gp import reduce_load_GP
        sage: reduce_load_GP()
        PARI/GP interpreter
    """
    return gp


def gp_console():
    """
    Spawn a new GP command-line session.

    EXAMPLES::

        sage: gp.console()  # not tested
        GP/PARI CALCULATOR Version 2.4.3 (development svn-12577)
        amd64 running linux (x86-64/GMP-4.2.1 kernel) 64-bit version
        compiled: Jul 21 2010, gcc-4.6.0 20100705 (experimental) (GCC)
        (readline v6.0 enabled, extended help enabled)
    """
    from sage.repl.rich_output.display_manager import get_display_manager
    if not get_display_manager().is_in_terminal():
        raise RuntimeError('Can use the console only in the terminal. '
                           'Try %%gp magics instead.')
    os.system('gp')


def gp_version():
    """
    EXAMPLES::

        sage: gp.version()  # not tested
        ((2, 4, 3), 'GP/PARI CALCULATOR Version 2.4.3 (development svn-12577)')
    """
    v = gp.eval(r'\v')
    i = v.find("Version ")
    w = v[i + len("Version "):]
    i = w.find(' ')
    w = w[:i]
    t = tuple([int(n) for n in w.split('.')])
    k = v.find('\n')
    return t, v[:k].strip()



r"""
Rational point sets on a Jacobian of a general hyperelliptic curve

AUTHORS:

- Sabrina Kunzweiler, Gareth Ma, Giacomo Pope (2024): adapt to smooth model
"""

# ****************************************************************************
#       Copyright (C) 2025 Sabrina Kunzweiler <sabrina.kunzweiler@math.u-bordeaux.fr>
#                     2025 Gareth Ma <grhkm21@gmail.com>
#                     2025 Giacomo Pope <giacomopope@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.functional import symbolic_prod as product
from sage.misc.prandom import choice
from sage.rings.finite_rings.finite_field_base import FiniteField as FiniteField_generic
from sage.rings.integer import Integer
from sage.rings.polynomial.polynomial_ring import polygen
from sage.schemes.generic.homset import SchemeHomset_points
from sage.schemes.weighted_projective.weighted_projective_point import (
    SchemeMorphism_point_weighted_projective_ring,
)
from sage.structure.element import parent


class HyperellipticJacobianHomset(SchemeHomset_points):
    r"""
    Set of rational points of the Jacobian.
    """

    def __init__(self, Y, X, **kwds):
        r"""
        Create the Hom-set of a Jacobian.

        The `K`-rational points of the Jacobian `J` over `k`
        are identified with the set of morphisms
        `\mathrm{Spec}(K) \to J`.

        INPUT:

        - ``Y`` -- domain (Spec(K) where K is the field of definition)
        - ``X`` -- codomain (the Jacobian)

        EXAMPLE::

            sage: from sage.schemes.hyperelliptic_curves.jacobian_homset_generic import HyperellipticJacobianHomset
            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(2*x^4 - x^3 + 4*x^2 - x, x^3 + x)
            sage: J = H.jacobian()
            sage: JQ = HyperellipticJacobianHomset(Spec(QQ), J); JQ
            Abelian group of points over Rational Field on Jacobian of Hyperelliptic Curve over Rational Field defined by y^2 + (x^3 + x)*y = 2*x^4 - x^3 + 4*x^2 - x
        """
        SchemeHomset_points.__init__(self, Y, X, **kwds)
        self._morphism_element = None

    def _repr_(self) -> str:
        r"""
        Return the string representation of the Jacobian Hom-set.

        EXAMPLES::

            sage: R.<x> = GF(13)[]
            sage: H = HyperellipticCurve(x^5 + 2*x + 1)
            sage: J = H.jacobian()
            sage: J(GF(13)) # indirect doctest
            Abelian group of points over Finite Field of size 13 on Jacobian of Hyperelliptic Curve over Finite Field of size 13 defined by y^2 = x^5 + 2*x + 1
        """
        return f"Abelian group of points over {self.domain().coordinate_ring()} on {self.codomain()}"

    def _morphism(self, *args, **kwds):
        r"""
        TODO
        """
        return self._morphism_element(*args, **kwds)

    def curve(self):
        r"""
        On input the set of `L`-rational points of a Jacobian `Jac(H)` defined over `K`,
        return the curve `H`.

        NOTE:
        The base field of `H` is not extended to `L`.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: K.<omega> = QQ.extension(x^2+x+1)
            sage: H = HyperellipticCurve(x^6-1)
            sage: JK = Jacobian(H)(K); JK
            Abelian group of points over Number Field in omega with defining polynomial x^2 + x + 1 on Jacobian of Hyperelliptic Curve over Rational Field defined by y^2 = x^6 - 1
            sage: JK.curve()
            Hyperelliptic Curve over Rational Field defined by y^2 = x^6 - 1
        """
        return self.codomain().curve()

    def extended_curve(self):
        r"""
        On input the set of `L`-rational points of a Jacobian `Jac(H)` defined over `K`,
        return the curve `H` with base extended to `L`.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: K.<omega> = QQ.extension(x^2+x+1)
            sage: H = HyperellipticCurve(x^6-1)
            sage: JK = Jacobian(H)(K); JK
            Abelian group of points over Number Field in omega with defining polynomial x^2 + x + 1 on Jacobian of Hyperelliptic Curve over Rational Field defined by y^2 = x^6 - 1
            sage: JK.extended_curve()
            Hyperelliptic Curve over Number Field in omega with defining polynomial x^2 + x + 1 defined by y^2 = x^6 - 1
        """
        # Code from schemes/generic/homset.py
        if "_extended_curve" in self.__dict__:
            return self._extended_curve
        R = self.domain().coordinate_ring()
        if R is not self.curve().base_ring():
            X = self.curve().base_extend(R)
        else:
            X = self.curve()
        self._extended_curve = X
        return X

    @cached_method
    def order(self) -> Integer:
        r"""
        Compute the order of the Jacobian.

        EXAMPLES:

        We compute the order of a superspecial hyperelliptic curve of genus 3::

            sage: R.<x> = GF(7)[]
            sage: H = HyperellipticCurve(x^8 - 1)
            sage: J = H.jacobian()
            sage: J(GF(7)).order() == (7+1)^3
            True
            sage: J(GF(7^2)).order() == (7+1)^6
            True
            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^8 - 1)
            sage: J = H.jacobian()
            sage: J.order()
            Traceback (most recent call last):
            ...
            NotImplementedError: order computation is only implemented for Jacobians over finite fields
        """
        if isinstance(self.base_ring(), FiniteField_generic):
            return sum(self.extended_curve().frobenius_polynomial())

        raise NotImplementedError(
            "order computation is only implemented for Jacobians over finite fields"
        )

    @cached_method
    def _curve_frobenius_roots(self):
        r"""
        Return the roots of the charpoly of frobenius on the extended curve.

        EXAMPLES:

        The following genus-2 curve is supersingular. The roots of Frobenius
        over `\mathbb{F}_5` are given by `\pm \sqrt{5}`, whereas over
        `\FF_{5^2}` Frobenius acts as multiplication by `-5`::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(x^6-1)
            sage: J = Jacobian(H)
            sage: rts = J(GF(5))._curve_frobenius_roots()
            sage: rts[0]^2 == -5
            True
            sage: J(GF(5^2))._curve_frobenius_roots()
            [-5, -5, -5, -5]
        """
        from sage.rings.qqbar import QQbar

        roots = self.extended_curve().frobenius_polynomial().roots(QQbar)
        return [r for r, e in roots for _ in range(e)]

    def cardinality(self, extension_degree=1) -> Integer:
        r"""
        Return `|Jac(C) / \mathbb{F}_{q^n}|`.

        EXAMPLES::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(x^6 + x + 1)
            sage: J = H.jacobian()
            sage: J(GF(5)).cardinality()
            31
            sage: J(GF(5^2)).cardinality()
            961
        """
        K = self.extended_curve().base_ring()
        if not isinstance(K, FiniteField_generic):
            raise NotImplementedError(
                "cardinality is only implemented for Jacobians over finite fields"
            )

        return Integer(
            product(1 - r**extension_degree for r in self._curve_frobenius_roots())
        )

    def count_points(self, n=1):
        r"""
        Count the number of points of the Jacobian over all finite extensions
        of the base fields of degree less than or equal to n.

        INPUT:

        - ``n`` -- a positive integer

        EXAMPLES::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(x^6 + x + 1)
            sage: J = H.jacobian()
            sage: J.count_points(10) == [J.change_ring(GF((5, k))).order() for k in range(1, 11)]
            True
            sage: J2 = J(GF((5, 2)))
            sage: J2.count_points(5) == J.count_points(10)[1::2]
            True
        """
        try:
            n = Integer(n)
        except TypeError:
            raise TypeError("n must be a positive integer")

        if n < 1:
            raise ValueError("n must be a positive integer")

        if n == 1:
            return self.cardinality()

        return [self.cardinality(extension_degree=i) for i in range(1, n + 1)]

    def point_to_mumford_coordinates(self, P):
        r"""
        On input a point P, return the Mumford coordinates
        of (the affine part of) the divisor [P].

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^5 - 2*x^4 + 2*x^3 - x^2, 1)
            sage: P = H([2,3]); P
            (2 : 3 : 1)
            sage: JQ = H.jacobian()(QQ)
            sage: JQ.point_to_mumford_coordinates(P)
            (x - 2, 3)
        """
        R, x = self.extended_curve().polynomial_ring().objgen()
        X, Y, Z = P._coords
        if Z == 0:
            return R.one(), R.zero()
        u = x - X
        v = R(Y)
        return u, v

    def _element_constructor_(self, *args, check=True):
        r"""
        Return a rational point in the abstract Homset `J(K)`, given:

        1. No arguments or the integer `0`; return `0 \in J`;

        2. A point `P` on `J = Jac(C)`, return `P`;

        3. A point `P` on the curve `H` such that `J = Jac(H)`;
           return `[P - P_0]`, where `P_0` is the distinguished point of `H`.
           By default, `P_0 = \infty`;

        4. Two points `P, Q` on the curve `H` such that `J = Jac(H)`;
           return `[P - Q]`;

        5. Polynomials `(u, v)` such that `v^2 + hv - f \equiv 0 \pmod u`;
           return `[(u(x), y - v(x))]`.

        EXAMPLES:

        First consider a hyperelliptic curve with an odd-degree model,
        hence a unique point at infinity::

            sage: R.<x> = GF(13)[]
            sage: H = HyperellipticCurve(x^7 + x + 1)
            sage: J = Jacobian(H)
            sage: JH = J.point_homset()
            sage: P = H.lift_x(1)
            sage: Q = H.lift_x(2)
            sage: D1 = JH(P); D1
            (x + 12, 4)
            sage: D2 = JH(Q); D2
            (x + 11, 1)
            sage: D = JH(P,Q); D
            (x^2 + 10*x + 2, 8*x + 9)
            sage: D == D1 - D2
            True
            sage: JH(x^2+10*x+2, 8*x+9) == D
            True

        In general, a distinguished point is used to embed points of the curve
        in the Jacobian. This works for general models of hyperelliptic curves::

            sage: R.<x> = PolynomialRing(GF(13))
            sage: H = HyperellipticCurve(2*x^8 + x + 1)
            sage: H.is_inert()
            True
            sage: J = Jacobian(H)
            sage: JH = J.point_homset()
            sage: P = H.lift_x(1)
            sage: D1 = JH(P); D1
            (x^2 + 12*x, 3*x + 12 : 0)

        To understand this output, one needs to look at the distinguished
        point::

            sage: P0 = H.distinguished_point(); P0
            (0 : 1 : 1)
            sage: JH(P) == JH(P,P0)
            True
            sage: JH(P0)
            (1, 0 : 1)

        We may change the distinguished point. Of course, the divisor `[P - Q]`
        does not depend on the choice of the distinguished point::

            sage: Q = H.lift_x(3)
            sage: JH(P,Q)
            (x^2 + 9*x + 3, 4*x + 11 : 0)
            sage: H.set_distinguished_point(H.random_point())
            sage: JH(P)  # random
            (x^2 + 6*x + 6, 10*x + 5 : 0)
            sage: JH(P,Q)
            (x^2 + 9*x + 3, 4*x + 11 : 0)

        TESTS:

        Ensure that field elements are treated as mumford coordinates::

            sage: R.<x> = GF(7)[]
            sage: H = HyperellipticCurve(x^7 - x^2 - 1)
            sage: J = H.jacobian(); J
            Jacobian of Hyperelliptic Curve over Finite Field of size 7 defined by y^2 = x^7 + 6*x^2 + 6
            sage: J(H.lift_x(3))
            (x + 4, 0)
            sage: _ == J(x + 4, 0) == J(x + 4, R(0))
            True

        Ensure that element constructor is user-friendly::

            sage: R.<x> = QQ[]
            sage: f = (x^4 - 2*x^2 - 8*x + 1) * (x^3 + x + 1)
            sage: H = HyperellipticCurve(f)
            sage: J = H.jacobian()
            sage: D = J(H(0, 1))
            sage: D.base_ring()
            Rational Field
            sage: J.change_ring(GF(13))(D).parent().base_ring()
            Finite Field of size 13
            sage: R.<y> = QQ[]
            sage: J(y^3 + y + 1, 0)
            (x^3 + x + 1, 0)

        TODO:

        - Allow sending a field element corresponding to the x-coordinate of a point?

        - Use ``__classcall__`` to sanitise input?

        - Add doctest for base extension
        """
        R = self.extended_curve().polynomial_ring()

        if len(args) == 0 or (len(args) == 1 and args[0] == ()):
            # this returns the incorrect result because subclasses like the inert case may implement
            # .zero differently
            # return self._morphism_element(self, R.one(), R.zero(), check=check)
            return self.zero(check=check)

        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]

        if len(args) == 1:
            P1 = args[0]
            if P1 == 0:
                return self.zero(check=check)
            if isinstance(P1, self._morphism_element):
                if parent(P1) is self:
                    return P1
                # may have to change polynomial ring etc.
                # this case will now be handled below.
                args = P1.uv()
            elif isinstance(P1, SchemeMorphism_point_weighted_projective_ring):
                args = args + (
                    self.extended_curve().distinguished_point(),
                )  # this case will now be handled below.
            else:
                raise ValueError(
                    "the input must consist of one or two points, or Mumford coordinates"
                )

        if len(args) == 2:
            P1 = args[0]
            P2 = args[1]
            if isinstance(
                P1, SchemeMorphism_point_weighted_projective_ring
            ) and isinstance(P2, SchemeMorphism_point_weighted_projective_ring):
                u1, v1 = self.point_to_mumford_coordinates(P1)
                P2_inv = self.extended_curve().hyperelliptic_involution(P2)
                u2, v2 = self.point_to_mumford_coordinates(P2_inv)
                u, v = self.cantor_composition(u1, v1, u2, v2)
            else:
                # We try to coerce input Mumford coordinates to polynomials
                try:
                    u = R(P1)
                    v = R(P2)
                except ValueError:
                    raise ValueError(
                        "the input must consist of one or two points, or Mumford coordinates"
                    )

        if len(args) > 2:
            raise ValueError("at most two arguments are allowed as input")

        return self._morphism_element(self, u, v, check=check)

    def zero(self, check=True):
        r"""
        Return the zero element of this jacobian homset.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^5 + 1)
            sage: JQ = H.jacobian()(QQ)
            sage: JQ.zero()
            (1, 0)
        """
        H = self.extended_curve()
        R = H.polynomial_ring()
        return self._morphism_element(self, R.one(), R.zero(), check=check)

    def __cantor_double_generic(self, u1, v1):
        r"""
        Efficient cantor composition for doubling an affine divisor

        Returns the Cantor composition of (u1, v1) with (u1, v1)
        together with the degree of the polynomial ``s`` which is
        needed for computing weights for the split and inert models.

        """
        f, h = self.extended_curve().hyperelliptic_polynomials()

        # New mumford coordinates
        if h.is_zero():
            s, _, e2 = u1.xgcd(v1 + v1)
            u3 = (u1 // s) ** 2
            v3 = v1 + e2 * (f - v1**2) // s
        else:
            s, _, e2 = u1.xgcd(v1 + v1 + h)
            u3 = (u1 // s) ** 2
            v3 = v1 + e2 * (f - v1 * h - v1**2) // s
        v3 = v3 % u3

        return u3, v3, s.degree()

    def _cantor_composition_generic(self, u1, v1, u2, v2):
        r"""
        Helper function for the Cantor composition algorithm.

        OUTPUT:

        The Cantor composition of ``(u1, v1)`` with ``(u2, v2)``,
        together with the degree of the polynomial ``s`` which is
        needed for computing the weights for the split and inert models.

        TESTS::

            sage: R.<x> = GF(13)[]
            sage: H = HyperellipticCurve(x^5 + 2*x + 1)
            sage: JK = H.jacobian()(GF(13))
            sage: (u1,v1) = (x^2 + 1, 10*x + 6)
            sage: (u2,v2) = (x + 5, R(8))
            sage: JK._cantor_composition_generic(u1,v1,u2,v2)
            (x^3 + 5*x^2 + x + 5, 9*x^2 + 10*x + 2, 0)

        """
        # Collect data from HyperellipticCurve
        H = self.extended_curve()
        f, h = H.hyperelliptic_polynomials()
        g = H.genus()

        # Ensure D1 and D2 are semi-reduced divisors
        assert v1.degree() < u1.degree() and v2.degree() < u2.degree(), (
            "The degree of bi must be smaller than ai"
        )
        assert u1.degree() <= 2 * g + 2 and u2.degree() <= 2 * g + 2, (
            f"The degree of ai must be smaller than 2g+2, {u1.degree()}, {u2.degree()}"
        )

        # Special case: duplication law
        if u1 == u2 and v1 == v2:
            return self.__cantor_double_generic(u1, v1)

        # Step One
        s0, _, e2 = u1.xgcd(u2)
        v1_m_v2 = v1 - v2

        # Special case: when gcd(u0, u1) == 1 we can
        # avoid many expensive steps as we have s = 1
        if s0.is_one():
            u3 = u1 * u2
            v3 = v2 + e2 * u2 * v1_m_v2
            v3 = v3 % u3
            return u3, v3, 0

        # Step Two
        w0 = v1 + v2 + h

        # Another special case, when w0 is zero we skip
        # a xgcd and can return early
        if w0.is_zero():
            u3 = (u1 * u2) // (s0**2)
            v3 = v2 + e2 * v1_m_v2 * (u2 // s0)
            v3 = v3 % u3
            return u3, v3, s0.degree()

        # Step Three
        s, c1, c2 = s0.xgcd(w0)
        u3 = (u1 * u2) // (s**2)
        v3 = v2 + (c1 * e2 * v1_m_v2 * u2 + c2 * (f - h * v2 - v2**2)) // s
        v3 = v3 % u3
        return u3, v3, s.degree()

    def _cantor_reduction_generic(self, u0, v0):
        r"""
        Helper function for the Cantor composition algorithm.

        OUTPUT:

        The reduced divisor of ``(u0, v0)``.

        TESTS::

            sage: R.<x> = GF(13)[]
            sage: H = HyperellipticCurve(x^5 + 2*x + 1)
            sage: JK = H.jacobian()(GF(13))
            sage: (u1,v1) = (x^2 + 1, 10*x + 6)
            sage: (u2,v2) = (x + 5, R(8))
            sage: (u3,v3,s) = JK._cantor_composition_generic(u1,v1,u2,v2)
            sage: JK._cantor_reduction_generic(u3,v3)
            (x^2 + 5*x + 2, 9*x + 3)
        """
        # Collect data from HyperellipticCurve
        H = self.extended_curve()
        f, h = H.hyperelliptic_polynomials()

        # Compute u' and v'
        u1 = ((v0**2 + h * v0 - f) // u0).monic()
        v1 = (-h - v0) % u1

        return u1, v1

    def cantor_composition(self, u1, v1, u2, v2):
        r"""
        Return the Cantor composition of ``(u1, v1)`` and ``(u2, v2)``.

        EXAMPLES::

            sage: R.<x> = GF(13)[]
            sage: H = HyperellipticCurve(x^7 + x^5 + x + 1)
            sage: JF = Jacobian(H).point_homset()
            sage: (u1, v1) = (x^3 + 4*x^2, 10*x^2 + 7*x + 1)
            sage: (u2, v2) = (x^3 + 8*x^2 + 11*x + 2, x^2 + 9*x + 10)
            sage: JF.cantor_composition(u1, v1, u2, v2)
            (x^6 + 12*x^5 + 4*x^4 + 7*x^3 + 8*x^2, 5*x^5 + 2*x^4 + 12*x^2 + 7*x + 1)
        """
        u3, v3, _ = self._cantor_composition_generic(u1, v1, u2, v2)
        return u3, v3

    def cantor_reduction(self, u0, v0):
        r"""
        Apply one reduction step of Cantor's algorithm to  ``(u0, v0)``.

        Note that, in general, several steps are necessary the
        representation of a reduced divisor.

        EXAMPLES::

            sage: R.<x> = GF(13)[]
            sage: H = HyperellipticCurve(x^7 + x^5 + x + 1)
            sage: g = H.genus()
            sage: JF = Jacobian(H).point_homset()
            sage: (u0, v0) = (x^6 + 12*x^5 + 4*x^4 + 7*x^3 + 8*x^2, 5*x^5 + 2*x^4 + 12*x^2 + 7*x + 1)
            sage: (u1, v1) = JF.cantor_reduction(u0, v0)
            sage: u1.degree() <= g
            False
            sage: (u2, v2) = JF.cantor_reduction(u1, v1)
            sage: u2.degree() <= g
            True

        Applying the reduction step to a reduced divisor might have unintended output,
        as is illustrated below.

            sage: (u3, v3) = JF.cantor_reduction(u2, v2)
            sage: u3.degree() >= g
            True
        """
        return self._cantor_reduction_generic(u0, v0)

    def lift_u(self, u, all=False):
        r"""
        Return one or all points with given `u`-coordinate.

        This method is deterministic: it returns the same data each time when
        called with the same `u`.

        Currently only implemented for Jacobians over a finite field.

        INPUT:

        * ``u`` -- an element of the base ring of the Jacobian

        * ``all`` -- boolean (default: ``False``); if ``True``, return a
          (possibly empty) list of all points with the given `u`-coordinate; if
          ``False``, return just one point, or raise a ``ValueError`` if there
          are none.

        OUTPUT:

        A point on this Jacobian.

        EXAMPLES::

            sage: R.<x> = GF(1993)[]
            sage: H = HyperellipticCurve(x^5 + x + 1)
            sage: J = H.jacobian()
            sage: P = J.lift_u(x^2 + 42*x + 270); P
            (x^2 + 42*x + 270, 1837*x + 838)
        """
        H = self.extended_curve()
        K = H.base_ring()
        g = H.genus()
        R = H.polynomial_ring()

        H_is_split = H.is_split()

        if u.is_zero():
            if H_is_split:
                if not all:
                    return self._morphism_element(self, R.one(), R.zero(), 0)
                return [
                    self._morphism_element(self, R.one(), R.zero(), n)
                    for n in range(g + 1)
                ]
            if not all:
                return self.zero()
            return [self.zero()]

        if not isinstance(K, FiniteField_generic) or K.degree() != 1:
            raise NotImplementedError(
                "lift_u is only implemented for Jacobians over a finite field of prime order"
            )

        f, h = H.hyperelliptic_polynomials()
        u1, v1 = R.one(), R.zero()

        u = R(u).monic()
        u_factors = u.factor()
        vss = []

        for x, e in u_factors:
            # Solve y^2 + hy - f = 0 mod x

            # TODO: is this the most efficient method? Maybe we should write
            # a helper function which computes y^2 + hy - f = 0 mod x which
            # properly handles trivial cases like when x is linear?
            K_ext = K.extension(modulus=x, names="a")
            y_ext = polygen(K_ext, "y_ext")
            h_ = K_ext(h % x)
            f_ = K_ext(f % x)
            vs = (y_ext**2 + h_ * y_ext - f_).roots(multiplicities=False)
            if len(vs) == 0 and not all:
                raise ValueError(f"no point with u-coordinate {u} on {self}")

            vss.append(vs)

        def postprocess_uv(u1, v1, n=None):
            # This function converts (u, v) into the point being returned, handling split cases
            if H.is_split():
                if n is None or not (0 <= n <= g - u1.degree()):
                    # this is an internal function
                    raise ValueError(
                        f"bug: n must be an integer between 0 and {g - u1.degree()}"
                    )
                return self._morphism_element(self, u1, v1, n, check=False)

            # We need to ensure the degree of u is even
            if H.is_inert():
                if u1.degree() % 2:
                    # TODO: make composition with distinguished_point its own function?
                    P0 = self.extended_curve().distinguished_point()
                    X0, Y0, _ = P0._coords
                    X = R.gen()  # TODO use better variable names in this function
                    _, h = self.extended_curve().hyperelliptic_polynomials()
                    u0 = X - X0
                    v0 = R(-Y0 - h(X0))
                    u1, v1, _ = self._cantor_composition_generic(u1, v1, u0, v0)
                assert not (u1.degree() % 2), f"{u1} must have even degree"

            return self._morphism_element(self, u1, v1, check=False)

        import itertools

        points = []
        for vv in itertools.product(*vss):
            u1, v1 = R.one(), R.zero()
            try:
                for (x, e), v in zip(u_factors, map(R, vv)):
                    for _ in range(e):
                        u1, v1, _ = self._cantor_composition_generic(u1, v1, x, v)
            # v is not rational, so we skip it
            except (ValueError, AttributeError):
                pass

            if not all:
                return postprocess_uv(u1, v1, n=0)

            if H_is_split:
                for n in range(g - u1.degree() + 1):
                    points.append(postprocess_uv(u1, v1, n=n))
            else:
                points.append(postprocess_uv(u1, v1))

        return points

    def _random_element_cover(self, degree=None):
        r"""
        Return a random element from the Jacobian.

        Distribution is not uniformly random, but returns the entire group.

        TESTS::

            sage: K = FiniteField(101)
            sage: R.<x> = K[]
            sage: H = HyperellipticCurve(x^7 + x)
            sage: JK = H.jacobian()(K)
            sage: JK._random_element_cover() # random
            (x^3 + 29*x^2 + 81*x + 66, 96*x^2 + 22*x + 32)

            sage: K = FiniteField(2)
            sage: R.<x> = K[]
            sage: H = HyperellipticCurve(x^5 + 1, x)
            sage: JK = H.jacobian()(K)
            sage: JK._random_element_cover() # random
            (x + 1, 1)

        The case "inert & odd genus" is not implemented (see :issue:`41985`)::

            sage: K = GF(101)
            sage: x = polygen(K)
            sage: f = 7*x^8 + 74*x^7 + 55*x^6 + 3*x^5 + 74*x^4 + 48*x^3 + 15*x^2 + 26*x + 97
            sage: HyperellipticCurve(f).jacobian()(K)._random_element_cover()
            Traceback (most recent call last):
            ...
            NotImplementedError: unable to perform arithmetic for inert models of odd genus; consider extending the base field to adjoin the points at infinity
        """
        H = self.extended_curve()
        R = H.polynomial_ring()
        g = H.genus()

        # For the inert case, the genus must be even
        if H.is_inert() and g % 2:
            raise NotImplementedError('unable to perform arithmetic for inert models of odd genus; '
                                      'consider extending the base field to adjoin the points at infinity')

        if degree is None:
            degree = (-1, g)

        while True:
            # TODO: i think we can skip this and simply ensure u
            #       is even degree with composition with the distinguished
            #       point?
            # if H.is_inert() and (u.degree() % 2) == 1:
            #     #TODO: better method to sample even degree polynomials
            #     continue
            u = R.random_element(degree=degree, monic=True)
            try:
                return choice(self.lift_u(u, all=True))
            # TODO: better handling rather than looping with try / except?
            except IndexError:
                pass

    def _random_element_rational(self):
        r"""
        Return a random element from the Jacobian. This algorithm is faster
        than :meth:`_random_element_cover` but is **NOT** surjective on the set
        of points.

        EXAMPLES::

            sage: R.<x> = GF(5)[]
            sage: f = x^5 + 2*x^4 + 4*x^3 + x^2 + 4*x + 3
            sage: H = HyperellipticCurve(f)
            sage: J = H.jacobian()
            sage: J.order()
            16
            sage: JH = J.point_homset()
            sage: len(set(JH._random_element_rational() for _ in range(300)))
            8

        TESTS:

        The case "inert & odd genus" is not implemented (see :issue:`41985`)::

            sage: K = GF(101)
            sage: x = polygen(K)
            sage: f = 7*x^8 + 74*x^7 + 55*x^6 + 3*x^5 + 74*x^4 + 48*x^3 + 15*x^2 + 26*x + 97
            sage: HyperellipticCurve(f).jacobian()(K)._random_element_rational()
            Traceback (most recent call last):
            ...
            NotImplementedError: unable to perform arithmetic for inert models of odd genus; consider extending the base field to adjoin the points at infinity
        """
        H = self.extended_curve()
        g = H.genus()

        # We randomly sample 2g + 1 points on the hyperelliptic curve
        points = [H.random_point() for _ in range(2 * g + 1)]

        # We create 2g + 1 divisors of the form (P) - infty
        divisors = [self(P) for P in points if P[2] != 0]

        # If we happened to only sample the point at infinity, we return this
        # Otherwise we compute the sum of all divisors.
        if not divisors:
            return self.zero()
        return sum(divisors, start=self.zero())

    def random_element(self, fast=False, *args, **kwargs):
        r"""
        Returns a random element from the Jacobian. Distribution is **NOT**
        uniformly random.

        INPUT:

        - ``fast`` -- (boolean, default ``True``) If set to ``True``, a fast
          algorithm is used, but the output is **NOT** guaranteed to cover the
          entire Jacobian. See examples below. If set to ``False``, a slower
          algorithm is used, but covers the entire Jacobian.

        EXAMPLES::

            sage: R.<x> = GF(5)[]
            sage: f = x^5 + 2*x^4 + 4*x^3 + x^2 + 4*x + 3
            sage: H = HyperellipticCurve(f)
            sage: J = H.jacobian()

        This example demonstrates that the ``fast`` algorithm is not
        necessarily surjective on the rational points of the Jacobian::

            sage: JH = J.point_homset()
            sage: len(set(JH.random_element(fast=True) for _ in range(300)))
            8
            sage: len(set(JH.random_element(fast=False) for _ in range(300)))
            16
            sage: J.order()
            16
        """
        if not isinstance(self.base_ring(), FiniteField_generic):
            raise NotImplementedError(
                "random element of Jacobian is only implemented over Finite Fields"
            )

        if fast:
            return self._random_element_rational(*args, **kwargs)
        return self._random_element_cover(*args, **kwargs)

    def points(self):
        r"""
        Return all points on this Jacobian `J(K)`.

        .. WARNING::

            This code is not efficient at all.

        EXAMPLES::

            sage: R.<x> = GF(3)[]
            sage: H = HyperellipticCurve(x^7 + 2*x + 1)
            sage: J3 = H.jacobian()(GF(3))
            sage: Pts = J3.points()
            sage: len(Pts)
            94
            sage: Pts[10]
            (x^2 + 2, x)
            sage: Pts[-1]
            (x^3 + 2*x^2 + 2*x + 1, 1)

        TESTS:

        The function also works in the split and inert cases.

            sage: R.<x> = GF(3)[]
            sage: H = HyperellipticCurve(x^8 + x + 2)
            sage: H.is_split()
            True
            sage: J3 = H.jacobian()(GF(3))
            sage: Pts = J3.points()
            sage: len(Pts)
            35

            sage: H = HyperellipticCurve(2*x^8 + 2*x + 1)
            sage: H.is_inert()
            True
            sage: J3 = H.jacobian()(GF(3))
            sage: Pts = J3.points(); len(Pts)
            29
        """
        H = self.extended_curve()
        R = H.polynomial_ring()
        g = H.genus()

        # TODO: after `monic` argument is added, use it
        ss = []
        for u in R.polynomials(max_degree=g):
            if u.is_monic():
                for P in self.lift_u(u, all=True):
                    # UGLY HACK: it keeps overcounting 0 without this...
                    # failing example: y^2 = x^5 + x over F_5
                    if not u.is_one() and list(P)[:2] == [1, 0]:
                        continue
                    ss.append(P)

        if H.is_ramified() or H.is_split():
            assert len(ss) == self.order()
            return ss

        # TODO: remove this
        # failing example: y^2 = 2x^6 + 1 over F_5
        ss = sorted(set(ss))
        assert len(ss) == self.order()
        return ss

    rational_points = points

    @cached_method
    def abelian_group(self):
        r"""
        Return the group of rational points on this Jacobian as an
        :class:`~sage.groups.additive_abelian.additive_abelian_wrapper.AdditiveAbelianGroupWrapper`
        object.

        EXAMPLES::

            sage: x = polygen(GF(419^2))
            sage: C = HyperellipticCurve(x^4 + x - 1)
            sage: C.is_split()
            True
            sage: C.jacobian().abelian_group()
            Additive abelian group isomorphic to Z/25193 + Z/7
              embedded in Abelian group of points over Finite Field in z2 of size 419^2
                on Jacobian of Hyperelliptic Curve over Finite Field in z2 of size 419^2 defined by y^2 = x^4 + x + 418

        ::

            sage: x = polygen(GF(419))
            sage: C = HyperellipticCurve(x^5 + x)
            sage: C.is_ramified()
            True
            sage: C.jacobian().abelian_group()
            Additive abelian group isomorphic to Z/29346 + Z/6
              embedded in Abelian group of points over Finite Field of size 419
                on Jacobian of Hyperelliptic Curve over Finite Field of size 419 defined by y^2 = x^5 + x

        ::

            sage: x = polygen(GF(419))
            sage: C = HyperellipticCurve(-x^6 + x - 3)
            sage: C.is_inert()
            True
            sage: C.jacobian().abelian_group()
            Additive abelian group isomorphic to Z/174078
              embedded in Abelian group of points over Finite Field of size 419
                on Jacobian of Hyperelliptic Curve over Finite Field of size 419 defined by y^2 = 418*x^6 + x + 416

        ::

            sage: x = polygen(GF(419))
            sage: C = HyperellipticCurve(11*x^6 + x)
            sage: C.is_inert()
            True
            sage: C.jacobian().abelian_group()
            Additive abelian group isomorphic to Z/420 + Z/420
              embedded in Abelian group of points over Finite Field of size 419
                on Jacobian of Hyperelliptic Curve over Finite Field of size 419 defined by y^2 = 11*x^6 + x

        ::

            sage: x = polygen(GF(419))
            sage: C = HyperellipticCurve(x^7 - x)
            sage: C.is_ramified()
            True
            sage: C.jacobian().abelian_group()
            Additive abelian group isomorphic to Z/420 + Z/420 + Z/210 + Z/2
              embedded in Abelian group of points over Finite Field of size 419
                on Jacobian of Hyperelliptic Curve over Finite Field of size 419 defined by y^2 = x^7 + 418*x

        ::

            sage: x = polygen(GF(419))
            sage: C = HyperellipticCurve(x^8 + x)
            sage: C.is_split()
            True
            sage: C.jacobian().abelian_group()
            Additive abelian group isomorphic to Z/420 + Z/420 + Z/420
              embedded in Abelian group of points over Finite Field of size 419
                on Jacobian of Hyperelliptic Curve over Finite Field of size 419 defined by y^2 = x^8 + x
        """
        n = self.order()
        g = self.curve().genus()
        from sage.groups.additive_abelian.additive_abelian_wrapper import AdditiveAbelianGroupWrapper
        A = AdditiveAbelianGroupWrapper(self, [], [])
        for fast in (True, False):
            for _ in range(99 * g):
                assert len(A.invariants()) <= 2 * g
                assert A.order().divides(n)
                if A.order() == n:
                    return A
                D = self.random_element(fast=fast)
                if D in A:
                    continue
                A = AdditiveAbelianGroupWrapper.from_generators(A._gen_elements + (D,))
        raise RuntimeError('very unlikely event, or (more likely) bug in HyperellipticJacobianHomset.abelian_group()')



"""
Root system data for (untwisted) type B affine
"""
#*****************************************************************************
#       Copyright (C) 2008-2009 Nicolas M. Thiery <nthiery at users.sf.net>,
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from .cartan_type import CartanType_standard_untwisted_affine


class CartanType(CartanType_standard_untwisted_affine):
    def __init__(self, n):
        """
        EXAMPLES::

            sage: ct = CartanType(['B',4,1])
            sage: ct
            ['B', 4, 1]
            sage: ct._repr_(compact = True)
            'B4~'

            sage: ct.is_irreducible()
            True
            sage: ct.is_finite()
            False
            sage: ct.is_affine()
            True
            sage: ct.is_untwisted_affine()
            True
            sage: ct.is_crystallographic()
            True
            sage: ct.is_simply_laced()
            False
            sage: ct.classical()
            ['B', 4]
            sage: ct.dual()
            ['B', 4, 1]^*
            sage: ct.dual().is_untwisted_affine()
            False

        TESTS::

            sage: TestSuite(ct).run()
        """
        assert n >= 1
        CartanType_standard_untwisted_affine.__init__(self, "B", n)

    def dynkin_diagram(self):
        """
        Return the extended Dynkin diagram for affine type `B`.

        EXAMPLES::

            sage: # needs sage.graphs
            sage: b = CartanType(['B',3,1]).dynkin_diagram(); b
                O 0
                |
                |
            O---O=>=O
            1   2   3
            B3~
            sage: b.edges(sort=True)
            [(0, 2, 1), (1, 2, 1), (2, 0, 1), (2, 1, 1), (2, 3, 2), (3, 2, 1)]
            sage: b = CartanType(['B',2,1]).dynkin_diagram(); b
            O=>=O=<=O
            0   2   1
            B2~
            sage: b.edges(sort=True)
            [(0, 2, 2), (1, 2, 2), (2, 0, 1), (2, 1, 1)]
            sage: b = CartanType(['B',1,1]).dynkin_diagram(); b
            O<=>O
            0   1
            B1~
            sage: b.edges(sort=True)
            [(0, 1, 2), (1, 0, 2)]
        """
        from . import cartan_type
        n = self.n
        if n == 1:
            res = cartan_type.CartanType(["A",1,1]).dynkin_diagram()
            res._cartan_type = self
            return res
        if n == 2:
            res = cartan_type.CartanType(["C",2,1]).relabel({0:0, 1:2, 2:1}).dynkin_diagram()
            res._cartan_type = self
            return res
        from .dynkin_diagram import DynkinDiagram_class
        g = DynkinDiagram_class(self)
        for i in range(1, n):
            g.add_edge(i, i+1)
        g.set_edge_label(n-1, n, 2)
        g.add_edge(0,2)
        return g

    def _latex_dynkin_diagram(self, label=None, node=None, node_dist=2, dual=False):
        r"""
        Return a latex representation of the Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['B',4,1])._latex_dynkin_diagram())
            \draw (0,0.7 cm) -- (2 cm,0);
            \draw (0,-0.7 cm) -- (2 cm,0);
            \draw (2 cm,0) -- (4 cm,0);
            \draw (4 cm, 0.1 cm) -- +(2 cm,0);
            \draw (4 cm, -0.1 cm) -- +(2 cm,0);
            \draw[shift={(5.2, 0)}, rotate=0] (135 : 0.45cm) -- (0,0) -- (-135 : 0.45cm);
            \draw[fill=white] (0 cm, 0.7 cm) circle (.25cm) node[left=3pt]{$0$};
            \draw[fill=white] (0 cm, -0.7 cm) circle (.25cm) node[left=3pt]{$1$};
            \draw[fill=white] (2 cm, 0 cm) circle (.25cm) node[below=4pt]{$2$};
            \draw[fill=white] (4 cm, 0 cm) circle (.25cm) node[below=4pt]{$3$};
            \draw[fill=white] (6 cm, 0 cm) circle (.25cm) node[below=4pt]{$4$};
            <BLANKLINE>

            sage: print(CartanType(['B',4,1]).dual()._latex_dynkin_diagram())
            \draw (0,0.7 cm) -- (2 cm,0);
            \draw (0,-0.7 cm) -- (2 cm,0);
            \draw (2 cm,0) -- (4 cm,0);
            \draw (4 cm, 0.1 cm) -- +(2 cm,0);
            \draw (4 cm, -0.1 cm) -- +(2 cm,0);
            \draw[shift={(4.8, 0)}, rotate=180] (135 : 0.45cm) -- (0,0) -- (-135 : 0.45cm);
            \draw[fill=white] (0 cm, 0.7 cm) circle (.25cm) node[left=3pt]{$0$};
            \draw[fill=white] (0 cm, -0.7 cm) circle (.25cm) node[left=3pt]{$1$};
            \draw[fill=white] (2 cm, 0 cm) circle (.25cm) node[below=4pt]{$2$};
            \draw[fill=white] (4 cm, 0 cm) circle (.25cm) node[below=4pt]{$3$};
            \draw[fill=white] (6 cm, 0 cm) circle (.25cm) node[below=4pt]{$4$};
            <BLANKLINE>
        """
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._latex_draw_node
        if self.n == 1:
            from . import cartan_type
            return cartan_type.CartanType(["A",1,1])._latex_dynkin_diagram(label, node, node_dist)
        if self.n == 2:
            from . import cartan_type
            return cartan_type.CartanType(["C",2,1])._latex_dynkin_diagram(label, node, node_dist, dual)
        n = self.n
        single_end = (n-2)*node_dist # Where the single line ends
        ret = "\\draw (0,0.7 cm) -- (%s cm,0);\n" % node_dist
        ret += "\\draw (0,-0.7 cm) -- (%s cm,0);\n" % node_dist
        ret += "\\draw (%s cm,0) -- (%s cm,0);\n" % (node_dist, single_end)
        ret += "\\draw (%s cm, 0.1 cm) -- +(%s cm,0);\n" % (single_end, node_dist)
        ret += "\\draw (%s cm, -0.1 cm) -- +(%s cm,0);\n" % (single_end, node_dist)
        if dual:
            ret += self._latex_draw_arrow_tip(single_end+0.5*node_dist-0.2, 0, 180)
        else:
            ret += self._latex_draw_arrow_tip(single_end+0.5*node_dist+0.2, 0, 0)
        ret += node(0, 0.7, label(0), 'left=3pt')
        ret += node(0, -0.7, label(1), 'left=3pt')
        for i in range(1, n):
            ret += node(i*node_dist, 0, label(i+1))
        return ret

    def ascii_art(self, label=None, node=None):
        """
        Return an ascii art representation of the extended Dynkin diagram.

        EXAMPLES::

            sage: print(CartanType(['B',3,1]).ascii_art())
                O 0
                |
                |
            O---O=>=O
            1   2   3

            sage: print(CartanType(['B',5,1]).ascii_art(label = lambda x: x+2))
                O 2
                |
                |
            O---O---O---O=>=O
            3   4   5   6   7

            sage: print(CartanType(['B',2,1]).ascii_art(label = lambda x: x+2))
            O=>=O=<=O
            2   4   3
            sage: print(CartanType(['B',1,1]).ascii_art(label = lambda x: x+2))
            O<=>O
            2   3
        """
        n = self.n
        from .cartan_type import CartanType
        if label is None:
            label = lambda i: i
        if node is None:
            node = self._ascii_art_node
        if n == 1:
            return CartanType(["A",1,1]).ascii_art(label, node)
        if n == 2:
            return CartanType(["C",2,1]).relabel({0:0, 1:2, 2:1}).ascii_art(label, node)
        ret = "    {} {}\n    |\n    |\n".format(node(label(0)), label(0))
        ret += "---".join(node(label(i)) for i in range(1,n)) + "=>={}\n".format(node(label(n)))
        ret += "".join("{!s:4}".format(label(i)) for i in range(1,n+1))
        return ret

    def _default_folded_cartan_type(self):
        """
        Return the default folded Cartan type.

        EXAMPLES::

            sage: CartanType(['B', 4, 1])._default_folded_cartan_type()
            ['B', 4, 1] as a folding of ['D', 5, 1]
        """
        from sage.combinat.root_system.type_folded import CartanTypeFolded
        n = self.n
        if n == 1:
            return CartanTypeFolded(self, ['A', 1, 1], [[0], [1]])
        return CartanTypeFolded(self, ['D', n + 1, 1],
            [[i] for i in range(n)] + [[n, n+1]])



r"""
Orthogonal Linear Groups

The general orthogonal group `GO(n,R)` consists of all `n \times n`
matrices over the ring `R` preserving an `n`-ary positive definite
quadratic form. In cases where there are multiple non-isomorphic
quadratic forms, additional data needs to be specified to
disambiguate. The special orthogonal group is the normal subgroup of
matrices of determinant one.

In characteristics different from 2, a quadratic form is equivalent to
a bilinear symmetric form. Furthermore, over the real numbers a
positive definite quadratic form is equivalent to the diagonal
quadratic form, equivalent to the bilinear symmetric form defined by
the identity matrix. Hence, the orthogonal group `GO(n,\RR)` is the
group of orthogonal matrices in the usual sense.

In the case of a finite field and if the degree `n` is even, then
there are two inequivalent quadratic forms and a third parameter ``e``
must be specified to disambiguate these two possibilities. The index
of `SO(e,d,q)` in `GO(e,d,q)` is `2` if `q` is odd, but `SO(e,d,q) =
GO(e,d,q)` if `q` is even.)

.. WARNING::

   GAP and Sage use different notations:

   * GAP notation: The optional ``e`` comes first, that is, ``GO([e,]
     d, q)``, ``SO([e,] d, q)``.

   * Sage notation: The optional ``e`` comes last, the standard Python
     convention: ``GO(d, GF(q), e=0)``, ``SO(d, GF(q), e=0)``.

EXAMPLES::

    sage: GO(3,7)
    General Orthogonal Group of degree 3 over Finite Field of size 7

    sage: G = SO(4, GF(7), 1); G
    Special Orthogonal Group of degree 4 and form parameter 1
     over Finite Field of size 7

    sage: # needs sage.libs.gap
    sage: G.random_element()   # random
    [4 3 5 2]
    [6 6 4 0]
    [0 4 6 0]
    [4 4 5 1]

TESTS::

    sage: G = GO(3, GF(5))
    sage: latex(G)
    \text{GO}_{3}(\Bold{F}_{5})
    sage: G = SO(3, GF(5))
    sage: latex(G)
    \text{SO}_{3}(\Bold{F}_{5})
    sage: G = SO(4, GF(5), 1)
    sage: latex(G)
    \text{SO}_{4}(\Bold{F}_{5}, +)

AUTHORS:

- David Joyner (2006-03): initial version

- David Joyner (2006-05): added examples, _latex_, __str__, gens,
  as_matrix_group

- William Stein (2006-12-09): rewrite

- Volker Braun (2013-1) port to new Parent, libGAP, extreme refactoring.

- Sebastian Oehms (2018-8) add
  :meth:`~sage.groups.matrix_gps.orthogonal.OrthogonalMatrixGroup_generic.invariant_form`
  (as alias), ``_OG``, option for user defined invariant bilinear form,
  and bug-fix in cmd-string for calling GAP (see :issue:`26028`)
"""

# ****************************************************************************
#       Copyright (C) 2006 David Joyner and William Stein
#                     2009 Mike Hansen
#                     2013 Volker Braun <vbraun.name@gmail.com>
#                     2018 Sebastian Oehms
#                     2018 Travis Scrimshaw
#                     2023 Matthias Koeppe
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.rings.integer_ring import ZZ
from sage.rings.finite_rings.finite_field_base import FiniteField
from sage.misc.latex import latex
from sage.misc.cachefunc import cached_method
from sage.groups.matrix_gps.named_group import (
    normalize_args_vectorspace, normalize_args_invariant_form,
    NamedMatrixGroup_generic)


def normalize_args_e(degree, ring, e):
    """
    Normalize the arguments that relate the choice of quadratic form
    for special orthogonal groups over finite fields.

    INPUT:

    - ``degree`` -- integer; the degree of the affine group, that is,
      the dimension of the affine space the group is acting on

    - ``ring`` -- a ring; the base ring of the affine space

    - ``e`` -- integer; one of `+1`, `0`, `-1`.  Only relevant for
      finite fields and if the degree is even. A parameter that
      distinguishes inequivalent invariant forms.

    OUTPUT: the integer ``e`` with values required by GAP

    TESTS::

        sage: from sage.groups.matrix_gps.orthogonal import normalize_args_e
        sage: normalize_args_e(2, GF(3), +1)
        1
        sage: normalize_args_e(3, GF(3), 0)
        0
        sage: normalize_args_e(3, GF(3), +1)
        0
        sage: normalize_args_e(2, GF(3), 0)
        Traceback (most recent call last):
        ...
        ValueError: must have e=-1 or e=1 for even degree
    """
    if isinstance(ring, FiniteField) and degree % 2 == 0:
        if e not in (-1, +1):
            raise ValueError('must have e=-1 or e=1 for even degree')
    else:
        e = 0
    return ZZ(e)


###############################################################################
# Orthogonal Group: common Code for both GO and SO
###############################################################################

def _OG(n, R, special, e=0, var='a', invariant_form=None):
    r"""
    This function is commonly used by the functions GO and SO to avoid
    unnecessarily duplicated code. For documentation and examples see
    the individual functions.

    TESTS:

    Check that :issue:`26028` is fixed::

        sage: GO(3,25).order()  # indirect doctest                                      # needs sage.libs.gap sage.rings.finite_rings
        31200

    Check that :issue:`28054` is fixed::

        sage: G = SO(2, GF(3), -1)
        sage: m = G.invariant_form()
        sage: G2 = SO(2, GF(3), 1, invariant_form=m)
        Traceback (most recent call last):
        ...
        NotImplementedError: invariant_form for finite groups is fixed by GAP
    """
    prefix = 'General'
    ltx_prefix = 'G'
    if special:
        prefix = 'Special'
        ltx_prefix = 'S'

    degree, ring = normalize_args_vectorspace(n, R, var=var)
    e = normalize_args_e(degree, ring, e)

    if invariant_form is not None:
        if isinstance(ring, FiniteField):
            raise NotImplementedError("invariant_form for finite groups is fixed by GAP")

    if e == 0:
        if invariant_form is not None:
            invariant_form = normalize_args_invariant_form(ring, degree, invariant_form)
            if not invariant_form.is_symmetric():
                raise ValueError("invariant_form must be symmetric")

            try:
                if invariant_form.is_positive_definite():
                    inserted_text = "with respect to positive definite symmetric form"
                else:
                    inserted_text = "with respect to non positive definite symmetric form"
            except ValueError:
                inserted_text = "with respect to symmetric form"

            name = '{0} Orthogonal Group of degree {1} over {2} {3}\n{4}'.format(
                prefix, degree, ring, inserted_text, invariant_form)
            ltx = r'\text{{{0}O}}_{{{1}}}({2})\text{{ {3} }}{4}'.format(
                ltx_prefix, degree, latex(ring), inserted_text,
                latex(invariant_form))
        else:
            name = '{0} Orthogonal Group of degree {1} over {2}'.format(prefix, degree, ring)
            ltx = r'\text{{{0}O}}_{{{1}}}({2})'.format(ltx_prefix, degree, latex(ring))
    else:
        name = '{0} Orthogonal Group of degree {1} and form parameter {2} over {3}'.format(prefix, degree, e, ring)
        ltx = r'\text{{{0}O}}_{{{1}}}({2}, {3})'.format(ltx_prefix, degree,
                                                        latex(ring),
                                                        '+' if e == 1 else '-')

    if isinstance(ring, FiniteField):
        try:
            from .orthogonal_gap import OrthogonalMatrixGroup_gap
        except ImportError:
            pass
        else:
            cmd = '{0}O({1}, {2}, {3})'.format(ltx_prefix, e, degree, ring.order())
            return OrthogonalMatrixGroup_gap(degree, ring, False, name, ltx, cmd)

    return OrthogonalMatrixGroup_generic(degree, ring, False, name, ltx, invariant_form=invariant_form)


########################################################################
# General Orthogonal Group
########################################################################

def GO(n, R, e=0, var='a', invariant_form=None):
    r"""
    Return the general orthogonal group.

    The general orthogonal group `GO(n,R)` consists of all `n \times n`
    matrices over the ring `R` preserving an `n`-ary positive definite
    quadratic form. In cases where there are multiple non-isomorphic
    quadratic forms, additional data needs to be specified to
    disambiguate.

    In the case of a finite field and if the degree `n` is even, then
    there are two inequivalent quadratic forms and a third parameter
    ``e`` must be specified to disambiguate these two possibilities.

    .. NOTE::

        This group is also available via ``groups.matrix.GO()``.

    INPUT:

    - ``n`` -- integer; the degree

    - ``R`` -- ring or an integer; if an integer is specified, the
      corresponding finite field is used

    - ``e`` -- ``+1`` or ``-1``, and ignored by default; only relevant
      for finite fields and if the degree is even: a parameter that
      distinguishes inequivalent invariant forms

    - ``var`` -- (default: ``'a'``) variable used to
      represent generator of the finite field, if needed

    - ``invariant_form`` -- (optional) instances being accepted by
      the matrix-constructor which define a `n \times n` square matrix
      over ``R`` describing the symmetric form to be kept invariant
      by the orthogonal group; the form is checked to be
      non-degenerate and symmetric but not to be positive definite

    OUTPUT: the general orthogonal group of given degree, base ring, and
    choice of invariant form

    EXAMPLES::

        sage: GO(3, GF(7))
        General Orthogonal Group of degree 3 over Finite Field of size 7

        sage: # needs sage.libs.gap
        sage: GO(3, GF(7)).order()
        672
        sage: GO(3, GF(7)).gens()
        (
        [3 0 0]  [0 1 0]
        [0 5 0]  [1 6 6]
        [0 0 1], [0 2 1]
        )

    Using the ``invariant_form`` option::

        sage: m = matrix(QQ, 3, 3, [[0, 1, 0], [1, 0, 0], [0, 0, 3]])
        sage: GO3  = GO(3, QQ)
        sage: GO3m = GO(3, QQ, invariant_form=m)
        sage: GO3 == GO3m
        False
        sage: GO3.invariant_form()
        [1 0 0]
        [0 1 0]
        [0 0 1]
        sage: GO3m.invariant_form()
        [0 1 0]
        [1 0 0]
        [0 0 3]
        sage: pm = Permutation([2,3,1]).to_matrix()
        sage: g = GO3(pm); g in GO3; g
        True
        [0 0 1]
        [1 0 0]
        [0 1 0]
        sage: GO3m(pm)
        Traceback (most recent call last):
        ...
        TypeError: matrix must be orthogonal with respect to the symmetric form
        [0 1 0]
        [1 0 0]
        [0 0 3]

        sage: GO(3,3, invariant_form=[[1,0,0], [0,2,0], [0,0,1]])
        Traceback (most recent call last):
        ...
        NotImplementedError: invariant_form for finite groups is fixed by GAP
        sage: 5 + 5
        10
        sage: R.<x> = ZZ[]
        sage: GO(2, R, invariant_form=[[x,0], [0,1]])
        General Orthogonal Group of degree 2 over
         Univariate Polynomial Ring in x over Integer Ring with respect to symmetric form
        [x 0]
        [0 1]

    TESTS::

        sage: # needs sage.libs.gap
        sage: TestSuite(GO3).run()
        sage: groups.matrix.GO(2, 3, e=-1)
        General Orthogonal Group of degree 2 and form parameter -1 over Finite Field of size 3
    """
    return _OG(n, R, False, e=e, var=var, invariant_form=invariant_form)


########################################################################
# Special Orthogonal Group
########################################################################

def SO(n, R, e=None, var='a', invariant_form=None):
    r"""
    Return the special orthogonal group.

    The special orthogonal group `GO(n,R)` consists of all `n \times n`
    matrices with determinant one over the ring `R` preserving an
    `n`-ary positive definite quadratic form. In cases where there are
    multiple non-isomorphic quadratic forms, additional data needs to
    be specified to disambiguate.

    .. NOTE::

        This group is also available via ``groups.matrix.SO()``.

    INPUT:

    - ``n`` -- integer; the degree

    - ``R`` -- ring or an integer; if an integer is specified, the
      corresponding finite field is used

    - ``e`` -- ``+1`` or ``-1``, and ignored by default; only relevant
      for finite fields and if the degree is even: a parameter that
      distinguishes inequivalent invariant forms

    - ``var`` -- (default: ``'a'``) variable used to
      represent generator of the finite field, if needed

    - ``invariant_form`` -- (optional) instances being accepted by
      the matrix-constructor which define a `n \times n` square matrix
      over ``R`` describing the symmetric form to be kept invariant
      by the orthogonal group; the form is checked to be
      non-degenerate and symmetric but not to be positive definite

    OUTPUT: the special orthogonal group of given degree, base ring, and choice
    of invariant form

    EXAMPLES::

        sage: G = SO(3,GF(5)); G
        Special Orthogonal Group of degree 3 over Finite Field of size 5

        sage: # needs sage.libs.gap
        sage: G = SO(3,GF(5))
        sage: G.gens()
        (
        [2 0 0]  [3 2 3]  [1 4 4]
        [0 3 0]  [0 2 0]  [4 0 0]
        [0 0 1], [0 3 1], [2 0 4]
        )
        sage: G.as_matrix_group()
        Matrix group over Finite Field of size 5 with 3 generators (
        [2 0 0]  [3 2 3]  [1 4 4]
        [0 3 0]  [0 2 0]  [4 0 0]
        [0 0 1], [0 3 1], [2 0 4]
        )

    Using the ``invariant_form`` option::

        sage: # needs sage.rings.number_field
        sage: CF3 = CyclotomicField(3); e3 = CF3.gen()
        sage: m = matrix(CF3, 3, 3, [[1,e3,0], [e3,2,0], [0,0,1]])
        sage: SO3  = SO(3, CF3)
        sage: SO3m = SO(3, CF3, invariant_form=m)
        sage: SO3 == SO3m
        False
        sage: SO3.invariant_form()
        [1 0 0]
        [0 1 0]
        [0 0 1]
        sage: SO3m.invariant_form()
        [    1 zeta3     0]
        [zeta3     2     0]
        [    0     0     1]
        sage: pm = Permutation([2,3,1]).to_matrix()
        sage: g = SO3(pm); g in SO3; g
        True
        [0 0 1]
        [1 0 0]
        [0 1 0]
        sage: SO3m(pm)
        Traceback (most recent call last):
        ...
        TypeError: matrix must be orthogonal with respect to the symmetric form
        [    1 zeta3     0]
        [zeta3     2     0]
        [    0     0     1]

        sage: SO(3, 5, invariant_form=[[1,0,0], [0,2,0], [0,0,3]])
        Traceback (most recent call last):
        ...
        NotImplementedError: invariant_form for finite groups is fixed by GAP
        sage: 5+5
        10

    TESTS::

        sage: TestSuite(SO3m).run()                                                     # needs sage.rings.number_field
        sage: groups.matrix.SO(2, 3, e=1)
        Special Orthogonal Group of degree 2 and form parameter 1 over Finite Field of size 3
    """
    return _OG(n, R, True, e=e, var=var, invariant_form=invariant_form)


########################################################################
# Orthogonal Group class
########################################################################

class OrthogonalMatrixGroup_generic(NamedMatrixGroup_generic):
    r"""
    General Orthogonal Group over arbitrary rings.

    EXAMPLES::

        sage: G = GO(3, GF(7)); G
        General Orthogonal Group of degree 3 over Finite Field of size 7
        sage: latex(G)
        \text{GO}_{3}(\Bold{F}_{7})

        sage: G = SO(3, GF(5));  G
        Special Orthogonal Group of degree 3 over Finite Field of size 5
        sage: latex(G)
        \text{SO}_{3}(\Bold{F}_{5})

        sage: # needs sage.rings.number_field
        sage: CF3 = CyclotomicField(3); e3 = CF3.gen()
        sage: m = matrix(CF3, 3,3, [[1,e3,0],[e3,2,0],[0,0,1]])
        sage: G = SO(3, CF3, invariant_form=m)
        sage: latex(G)
        \text{SO}_{3}(\Bold{Q}(\zeta_{3}))\text{ with respect to non positive definite symmetric form }\left(\begin{array}{rrr}
        1 & \zeta_{3} & 0 \\
        \zeta_{3} & 2 & 0 \\
        0 & 0 & 1
        \end{array}\right)
    """

    @cached_method
    def invariant_bilinear_form(self):
        """
        Return the symmetric bilinear form preserved by ``self``.

        OUTPUT: a matrix

        EXAMPLES::

            sage: # needs sage.libs.gap
            sage: GO(2,3,+1).invariant_bilinear_form()
            [0 1]
            [1 0]
            sage: GO(2,3,-1).invariant_bilinear_form()
            [2 1]
            [1 1]
            sage: G = GO(4, QQ)
            sage: G.invariant_bilinear_form()
            [1 0 0 0]
            [0 1 0 0]
            [0 0 1 0]
            [0 0 0 1]
            sage: GO3m = GO(3, QQ, invariant_form=(1,0,0, 0,2,0, 0,0,3))
            sage: GO3m.invariant_bilinear_form()
            [1 0 0]
            [0 2 0]
            [0 0 3]

        TESTS::

            sage: GO3m.invariant_form()                                                 # needs sage.libs.gap
            [1 0 0]
            [0 2 0]
            [0 0 3]
        """
        if self._invariant_form is not None:
            return self._invariant_form

        from sage.matrix.constructor import identity_matrix
        m = identity_matrix(self.base_ring(), self.degree())
        m.set_immutable()
        return m

    invariant_quadratic_form = invariant_bilinear_form
    # this is identical in the generic case

    invariant_form = invariant_bilinear_form
    # alias (analogues to symplectic and unitary cases)

    def _check_matrix(self, x, *args):
        """
        Check whether the matrix ``x`` is orthogonal.

        See :meth:`~sage.groups.matrix_gps.matrix_group._check_matrix`
        for details.

        EXAMPLES::

            sage: G = GO(4, GF(5), +1)
            sage: G._check_matrix(G.an_element().matrix())
        """
        if self._special and x.determinant() != 1:
            raise TypeError('matrix must have determinant one')
        F = self.invariant_bilinear_form()
        if x * F * x.transpose() != F:
            if F == self.one().matrix():
                raise TypeError('matrix must be orthogonal')
            else:
                raise TypeError('matrix must be orthogonal with respect to the symmetric form\n%s' % (F))
        # TODO: check that quadratic form is preserved in characteristic two



r"""
Elements lying in extension of rings

AUTHOR:

- Xavier Caruso (2019)
"""

# ###########################################################################
#    Copyright (C) 2019 Xavier Caruso <xavier.caruso@normalesup.org>
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 2 of the License, or
#    (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ***************************************************************************


cimport cython
from sage.ext.stdsage cimport PY_NEW
from sage.cpython.getattr cimport AttributeErrorMessage
from sage.cpython.getattr import dir_with_other_class
from sage.misc.latex import latex

from sage.structure.category_object import normalize_names
from sage.structure.element cimport CommutativeAlgebraElement
from sage.structure.parent cimport Parent
from sage.rings.integer_ring import ZZ
from sage.categories.fields import Fields
from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing

from sage.rings.ring_extension cimport RingExtension_generic, RingExtensionWithGen, RingExtensionFractionField
from sage.rings.ring_extension_morphism cimport MapRelativeRingToFreeModule, are_equal_morphisms
from sage.rings.ring_extension_conversion cimport backend_parent, backend_element
from sage.rings.ring_extension_conversion cimport to_backend, from_backend


# Classes
#########

cdef class RingExtensionElement(CommutativeAlgebraElement):
    r"""
    Generic class for elements lying in ring extensions.

    TESTS::

        sage: K = GF(5^4).over()                                                        # needs sage.rings.finite_rings
        sage: x = K.random_element()                                                    # needs sage.rings.finite_rings
        sage: TestSuite(x).run()                                                        # needs sage.rings.finite_rings
    """
    def __init__(self, RingExtension_generic parent, x, *args, **kwds):
        r"""
        Initialize this element.

        INPUT:

        - ``parent`` -- the parent of this element

        - ``x`` -- some data to construct this element

        TESTS::

            sage: Q = QQ.over(ZZ)
            sage: x = Q(1/2)
            sage: x
            1/2
        """
        if not isinstance(parent, RingExtension_generic):
            raise TypeError("%s is not a ring extension" % parent)
        x = backend_element(x)
        try:
            parentx = x.parent()
            if parent._base.has_coerce_map_from(parentx):
                x = parent._base.coerce_map_from(parentx)(x)
                x = parent._backend_defining_morphism(x)
        except AttributeError:
            pass
        CommutativeAlgebraElement.__init__(self, parent)
        ring = parent._backend
        self._backend = ring(x, *args, **kwds)

    def __reduce__(self):
        """
        Return a tuple of a function and data that can be used to unpickle this
        element.

        TESTS::

            sage: K = GF(5^3).over()
            sage: x = K.random_element()
            sage: type(x)
            <class 'sage.rings.ring_extension_element.RingExtensionWithBasisElement'>
            sage: loads(dumps(x)) == x
            True
        """
        return self._parent, (self._backend,)

    def __getattr__(self, name):
        """
        If the parent of this element was created with ``import_methods = True``,
        return a wrapper to the corresponding method of the backend element
        (if it exists).

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: A.<a> = QQ.extension(x^2 - 2)
            sage: K.<a> = A.over()  # over QQ
            sage: hasattr(a, 'continued_fraction')
            True
            sage: a.continued_fraction()
            [1; (2)*]
        """
        try:
            return self.getattr_from_category(name)
        except AttributeError:
            pass
        method = None
        if (<RingExtension_generic>self._parent)._import_methods and hasattr(self._backend, name):
            method = getattr(self._backend, name)
        if not callable(method):
            raise AttributeError(AttributeErrorMessage(self, name))

        def wrapper(*args, **kwargs):
            output = method(*to_backend(args), **to_backend(kwargs))
            return from_backend(output, self._parent)
        wrapper.__doc__ = method.__doc__
        return wrapper

    def __getitem__(self, i):
        r"""
        Return the `i`-th item of this element.

        This methods calls the appropriate method of the backend if
        ``import_methods`` is set to ``True``

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: E = R.over()
            sage: P = E(x^2 + 2*x + 3)
            sage: P[0]
            3
        """
        if (<RingExtension_generic>self._parent)._import_methods:
            output = self._backend[to_backend(i)]
            return from_backend(output, self._parent)
        return TypeError("this element is not subscriptable")

    def __call__(self, *args, **kwargs):
        r"""
        Call this element.

        This methods calls the appropriate method of the backend if
        ``import_methods`` is set to ``True``

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: E = R.over()
            sage: P = E(x^2 + 2*x + 3)
            sage: P(1)
            6
        """
        if (<RingExtension_generic>self._parent)._import_methods:
            output = self._backend(*to_backend(args), **to_backend(kwargs))
            if args:
                E = args[0].parent()
                if isinstance(E, RingExtension_generic):
                    return from_backend(output, args[0].parent())
                return output
            else:
                return from_backend(output, self._parent)
        return TypeError("this element is not callable")

    def __dir__(self):
        """
        Return the list of all the attributes of this element;
        if the parent of this element was created with ``import_methods = True``,
        concatenate this list with the list of all the methods of the backend
        element.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: A.<a> = QQ.extension(x^2 - 2)
            sage: K.<a> = A.over()
            sage: dir(a)
            ['__abs__',
             '__add__',
             ...
             'complex_embeddings',
             'conjugate',
             'continued_fraction',
             'continued_fraction_list',
             ...
             'trace',
             'valuation',
             'vector',
             'xgcd']
        """
        d = dir_with_other_class(self, self._parent.category().element_class)
        if not (<RingExtension_generic>self._parent)._import_methods:
            return d
        for name in dir(self._backend):
            try:
                attribute = getattr(self._backend, name)
                if callable(attribute):
                    d.append(name)
            except AttributeError:
                pass
        return sorted(set(d))

    def __hash__(self):
        """
        Return a hash of this element.

        EXAMPLES:

            sage: E.<a> = GF(5^3).over()                                                # needs sage.rings.finite_rings
            sage: hash(a)                                                               # needs sage.rings.finite_rings
            5
        """
        return hash(self._backend)

    def _repr_(self, **options):
        r"""
        Return a string representation of this element.

        Do not override this method in subclasses;
        instead override the method :meth:`_repr_extension`.

        TESTS::

            sage: K.<a> = GF(5^2).over()                                                # needs sage.rings.finite_rings
            sage: L.<b> = GF(5^4).over(K)                                               # needs sage.rings.finite_rings
            sage: b._repr_()                                                            # needs sage.rings.finite_rings
            'b'
        """
        cdef RingExtension_generic parent = self._parent
        if 'print_elements_as' in options:
            print_as = options.pop('print_elements_as')
        else:
            print_as = parent._print_options.get('print_elements_as')
        if print_as is not None:
            return print_as(self._backend)._repr_(**options)
        print_options = parent._print_options.copy()
        for (name, value) in options.items():
            method = None
            if hasattr(parent, '_print_option_' + name):
                method = getattr(parent, '_print_option_' + name)
            if not callable(method):
                raise ValueError("option '%s' does not exist" % name)
            print_options[name] = method(value)
        return self._repr_extension(**print_options)

    def _repr_extension(self, **options):
        r"""
        Return a string representation of this element.

        TESTS::

            sage: K = QQ.over(ZZ)
            sage: x = K(1/2)
            sage: x._repr_extension()
            '1/2'
        """
        return str(self._backend)

    def _latex_(self, **options):
        r"""
        Return a LaTeX representation of this element.

        Do not override this method in subclasses;
        instead override the method :meth:`_latex_extension`.

        TESTS::

            sage: K.<a> = GF(5^2).over()                                                # needs sage.rings.finite_rings
            sage: L.<b> = GF(5^4).over(K)                                               # needs sage.rings.finite_rings
            sage: b._latex_()                                                           # needs sage.rings.finite_rings
            'b'
        """
        cdef RingExtension_generic parent = self._parent
        if 'print_elements_as' in options:
            print_as = options.pop('print_elements_as')
        else:
            print_as = parent._print_options.get('print_elements_as')
        if print_as is not None:
            return print_as(self._backend)._latex_(**options)
        print_options = parent._print_options.copy()
        for (name, value) in options.items():
            method = None
            if hasattr(parent, '_print_option_' + name):
                method = getattr(parent, '_print_option_' + name)
            if not callable(method):
                raise ValueError("option '%s' does not exist" % name)
            print_options[name] = method(value)
        return self._latex_extension(**print_options)

    def _latex_extension(self, **options):
        r"""
        Return a LaTeX representation of this element.

        TESTS::

            sage: K = QQ.over(ZZ)
            sage: x = K(1/2)
            sage: x._latex_extension()
            \frac{1}{2}
        """
        return latex(self._backend)

    def backend(self, force=False):
        """
        Return the backend of this element.

        INPUT:

        - ``force`` -- boolean (default: ``False``); if ``False``,
          raise an error if the backend is not exposed

        EXAMPLES::

            sage: F = GF(5^2)
            sage: K.<z> = GF(5^4).over(F)
            sage: x = z^10
            sage: x
            (z2 + 2) + (3*z2 + 1)*z
            sage: y = x.backend()
            sage: y
            4*z4^3 + 2*z4^2 + 4*z4 + 4
            sage: y.parent()
            Finite Field in z4 of size 5^4
        """
        if force or (<RingExtension_generic>(self._parent))._is_backend_exposed:
            return self._backend
        raise ValueError("backend is not exposed; try force=True")

    def in_base(self):
        r"""
        Return this element as an element of the base.

        EXAMPLES::

            sage: F = GF(5^2)
            sage: K.<z> = GF(5^4).over(F)
            sage: x = z^3 + z^2 + z + 4
            sage: y = x.in_base()
            sage: y
            z2 + 1
            sage: y.parent()
            Finite Field in z2 of size 5^2

        When the element is not in the base, an error is raised::

            sage: z.in_base()                                                           # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            ValueError: z is not in the base

        ::

            sage: S.<X> = F[]
            sage: E = S.over(F)
            sage: f = E(1)
            sage: g = f.in_base(); g
            1
            sage: g.parent()
            Finite Field in z2 of size 5^2

        TESTS:

        We check the case of a tower of extensions::

            sage: F = GF(5^2)
            sage: K.<u> = GF(5^4).over(F)
            sage: L.<v> = GF(5^8).over(K)
            sage: x = 4*v^7 + v^6 + 3*v^4 + v^3 + v^2 + 4
            sage: x.in_base()
            u
        """
        cdef RingExtension_generic parent = <RingExtension_generic>self._parent
        if isinstance(parent, RingExtensionWithGen):
            v = self.vector()
            for i in range(1, len(v)):
                if v[i]:
                    raise ValueError("%s is not in the base" % self)
            return v[0]
        else:
            f = parent._backend_defining_morphism
            base = f.domain()
            ring = f.codomain()
            if ring.has_coerce_map_from(base) and are_equal_morphisms(f, None):
                return parent.base()(base(self._backend))
        raise NotImplementedError("cannot cast %s to the base" % self)

    cpdef _richcmp_(left, right, int op):
        r"""
        Compare this element with ``right`` according to
        the rich comparison operator ``op``.

        The comparison is performed by comparing the backend
        elements.

        INPUT:

        - ``right`` -- an element in the same parent

        - ``op`` -- the comparison operator

        EXAMPLES::

            sage: K.<a> = GF(5^2).over()
            sage: x = K.random_element()
            sage: x == x
            True
            sage: x == x + 1
            False
            sage: x == x^25
            True
        """
        return left._backend._richcmp_(backend_element(right), op)

    cpdef _add_(self, other):
        r"""
        Return the sum of this element and ``other``.

        TESTS::

            sage: K = GF(5^4).over(GF(5^2))
            sage: x = K.random_element()
            sage: y = K.random_element()
            sage: (x+y).parent() is K
            True
            sage: x + y == y + x
            True
        """
        cdef RingExtensionElement ans = PY_NEW(type(self))
        ans._parent = self._parent
        ans._backend = self._backend + (<RingExtensionElement>other)._backend
        return ans

    cpdef _neg_(self):
        r"""
        Return the opposite of this element.

        TESTS::

            sage: K = GF(5^4).over(GF(5^2))
            sage: x = K.random_element()
            sage: y = -x
            sage: y.parent() is K
            True
            sage: x + y == 0
            True
        """
        cdef RingExtensionElement ans = PY_NEW(type(self))
        ans._parent = self._parent
        ans._backend = -self._backend
        return ans

    cpdef _sub_(self, other):
        r"""
        Return the difference of this element and ``other``.

        TESTS::

            sage: K = GF(5^4).over(GF(5^2))
            sage: x = K.random_element()
            sage: y = K.random_element()
            sage: (x - y).parent() is K
            True
            sage: x - y == x + (-y)
            True
        """
        cdef RingExtensionElement ans = PY_NEW(type(self))
        ans._parent = self._parent
        ans._backend = self._backend - (<RingExtensionElement>other)._backend
        return ans

    cpdef _mul_(self, other):
        r"""
        Return the product of this element and ``other``.

        TESTS::

            sage: K = GF(5^4).over(GF(5^2))
            sage: x = K.random_element()
            sage: y = K.random_element()
            sage: (x*y).parent() is K
            True
            sage: x * y == y * x
            True
        """
        cdef RingExtensionElement ans = PY_NEW(type(self))
        ans._parent = self._parent
        ans._backend = self._backend * (<RingExtensionElement>other)._backend
        return ans

    cpdef _div_(self, other):
        r"""
        Return the quotient of this element by ``other``,
        considered as an element of the fraction field.

        TESTS::

            sage: x = polygen(ZZ, 'x')
            sage: A.<a> = ZZ.extension(x^2 - 2)
            sage: OK = A.over()
            sage: a = OK(a)
            sage: b = 1/a; b
            a/2
            sage: b.parent()
            Fraction Field of
             Maximal Order generated by a in Number Field in a
             with defining polynomial x^2 - 2 over its base
            sage: a*b
            1
        """
        cdef RingExtensionElement ans
        cdef RingExtension_generic parent = self._parent
        if parent._fraction_field is None:
            parent._fraction_field = parent.fraction_field()
            parent._fraction_field_type = <type>parent._fraction_field.element_class
        ans = PY_NEW(parent._fraction_field_type)
        ans._parent = parent._fraction_field
        ans._backend = self._backend / (<RingExtensionElement>other)._backend
        return ans

    def additive_order(self):
        r"""
        Return the additive order of this element.

        EXAMPLES::

            sage: K.<a> = GF(5^4).over(GF(5^2))                                         # needs sage.rings.finite_rings
            sage: a.additive_order()                                                    # needs sage.rings.finite_rings
            5
        """
        return self._backend.additive_order()

    def multiplicative_order(self):
        r"""
        Return the multiplicite order of this element.

        EXAMPLES::

            sage: K.<a> = GF(5^4).over(GF(5^2))                                         # needs sage.rings.finite_rings
            sage: a.multiplicative_order()                                              # needs sage.rings.finite_rings
            624
        """
        return self._backend.multiplicative_order()

    def is_unit(self):
        r"""
        Return whether if this element is a unit in this ring.

        EXAMPLES::

            sage: A.<x> = PolynomialRing(QQ)
            sage: E = A.over(QQ)
            sage: E(4).is_unit()
            True
            sage: E(x).is_unit()
            False
        """
        return self._backend.is_unit()

    def is_nilpotent(self):
        r"""
        Return whether if this element is nilpotent in this ring.

        EXAMPLES::

            sage: A.<x> = PolynomialRing(QQ)
            sage: E = A.over(QQ)
            sage: E(0).is_nilpotent()
            True
            sage: E(x).is_nilpotent()
            False
        """
        return self._backend.is_nilpotent()

    def is_prime(self):
        r"""
        Return whether this element is a prime element in this ring.

        EXAMPLES::

            sage: A.<x> = PolynomialRing(QQ)
            sage: E = A.over(QQ)
            sage: E(x^2 + 1).is_prime()                                                 # needs sage.libs.pari
            True
            sage: E(x^2 - 1).is_prime()                                                 # needs sage.libs.pari
            False
        """
        return self._backend.is_prime()

    def is_square(self, root=False):
        r"""
        Return whether this element is a square in this ring.

        INPUT:

        - ``root`` -- boolean (default: ``False``); if ``True``,
          return also a square root

        EXAMPLES::

            sage: K.<a> = GF(5^3).over()
            sage: a.is_square()
            False
            sage: a.is_square(root=True)
            (False, None)
            sage: b = a + 1
            sage: b.is_square()
            True
            sage: b.is_square(root=True)
            (True, 2 + 3*a + a^2)
        """
        is_sq = self._backend.is_square()
        sq = None
        if root and is_sq:
            sq = self.sqrt(extend=False, all=False)
        if root:
            return is_sq, sq
        else:
            return is_sq

    def sqrt(self, extend=True, all=False, name=None):
        r"""
        Return a square root or all square roots of this element.

        INPUT:

        - ``extend`` -- boolean (default: ``True``); if ``True``,
          return a square root in an extension ring, if necessary.
          Otherwise, raise a :exc:`ValueError` if the root is not in
          the ring.

        - ``all`` -- boolean (default: ``False``); if ``True``,
          return all square roots of this element, instead of just one

        - ``name`` -- required when ``extend=True`` and ``self`` is not a
          square; this will be the name of the generator extension

        .. NOTE::

            The option ``extend=True`` is often not implemented.

        EXAMPLES::

            sage: K.<a> = GF(5^3).over()
            sage: b = a + 1
            sage: b.sqrt()
            2 + 3*a + a^2
            sage: b.sqrt(all=True)
            [2 + 3*a + a^2, 3 + 2*a - a^2]
        """
        sq = self._backend.sqrt(extend=extend, all=all)
        if all:
            gen = sq[0]
        else:
            gen = sq
        parent = self._parent
        backend_parent = gen.parent()
        if backend_parent is not (<RingExtension_generic>parent)._backend:
            from sage.rings.ring_extension import RingExtension
            if name is None:
                raise ValueError("you must specify a variable name")
            names = normalize_names(1, name)
            constructor = (RingExtensionWithGen,
                           {'gen': gen, 'name': names[0], 'is_backend_exposed': False})
            parent = RingExtension(backend_parent, parent, (gen,), names, constructors=[constructor])
        if all:
            return [ parent(s) for s in sq ]
        else:
            return parent(sq)


# Fraction fields
#################

cdef class RingExtensionFractionFieldElement(RingExtensionElement):
    r"""
    A class for elements lying in fraction fields of ring extensions.

    TESTS::

        sage: Z = ZZ.over()
        sage: Q = Z.fraction_field()
        sage: x = Q.random_element()
        sage: type(x)
        <class 'sage.rings.ring_extension_element.RingExtensionFractionFieldElement'>
        sage: TestSuite(x).run()
    """
    def __hash__(self):
        """
        Return a hash of this element.

        EXAMPLES:

            sage: E.<a> = GF(5^3).over()                                                # needs sage.rings.finite_rings
            sage: hash(a)                                                               # needs sage.rings.finite_rings
            5
        """
        return hash(self._backend)

    def _repr_extension(self, **options):
        r"""
        Return a string representation of this element.

        TESTS::

            sage: Z = ZZ.over()
            sage: Q = Z.fraction_field()
            sage: x = Q(1/2)
            sage: x._repr_extension()
            '1/2'
            sage: R = QQ['x'].over()
            sage: K = R.fraction_field()
            sage: x = R.gen()
            sage: (x^2 + 1) / (x^2 - 1)
            (x^2 + 1)/(x^2 - 1)
            sage: x / (x + 1)
            x/(x + 1)
            sage: (x + 1)/(-x)
            (-x - 1)/x
        """
        num = self.numerator()
        denom = self.denominator()
        if denom == 1:
            sd = ""
        elif denom == -1:
            num = -num
            sd = ""
        elif denom._is_atomic():
            sd = "/%s" % denom
        elif (-denom)._is_atomic():
            sd = "/%s" % (-denom)
            num = -num
        else:
            sd = "/(%s)" % denom
        if num._is_atomic():
            return "%s%s" % (num, sd)
        else:
            return "(%s)%s" % (num, sd)

    def _latex_extension(self, **options):
        r"""
        Return a LaTeX representation of this element.

        TESTS::

            sage: Z = ZZ.over()
            sage: Q = Z.fraction_field()
            sage: x = Q(1/2)
            sage: x._latex_extension()
            '\\frac{1}{2}'
        """
        num = self.numerator()
        denom = self.denominator()
        if denom == -1:
            denom = 1
            num = -num
        if isinstance((<RingExtensionFractionField>self._parent)._ring, RingExtension_generic):
            snum = num._latex_(**options)
            sdenom = denom._latex_(**options)
        else:
            snum = latex(num)
            sdenom = latex(denom)
        if denom == 1:
            return snum
        else:
            return "\\frac{%s}{%s}" % (snum, sdenom)

    def numerator(self):
        r"""
        Return the numerator of this element.

        EXAMPLES::

            sage: x = polygen(ZZ, 'x')
            sage: A.<a> = ZZ.extension(x^2 - 2)
            sage: OK = A.over()  # over ZZ
            sage: K = OK.fraction_field(); K
            Fraction Field of Maximal Order generated by a in Number Field in a
             with defining polynomial x^2 - 2 over its base
            sage: x = K(1/a); x
            a/2
            sage: num = x.numerator(); num
            a

        The numerator is an element of the ring which was used
        to construct the fraction field::

            sage: num.parent()                                                          # needs sage.rings.number_field
            Maximal Order generated by a in Number Field in a
             with defining polynomial x^2 - 2 over its base
            sage: num.parent() is OK                                                    # needs sage.rings.number_field
            True

        TESTS::

            sage: x = K.random_element()
            sage: x == x.numerator() / x.denominator()
            True
        """
        ring = (<RingExtensionFractionField>self._parent)._ring
        num = self._backend.numerator()
        return ring(num)

    def denominator(self):
        r"""
        Return the denominator of this element.

        EXAMPLES::

            sage: R.<x> = ZZ[]
            sage: A.<a> = ZZ.extension(x^2 - 2)
            sage: OK = A.over()  # over ZZ
            sage: K = OK.fraction_field(); K
            Fraction Field of
             Maximal Order generated by a in Number Field in a
             with defining polynomial x^2 - 2 over its base
            sage: x = K(1/a); x
            a/2
            sage: denom = x.denominator(); denom
            2

        The denominator is an element of the ring which was used
        to construct the fraction field::

            sage: denom.parent()                                                        # needs sage.rings.number_field
            Maximal Order generated by a in Number Field in a with defining polynomial x^2 - 2 over its base
            sage: denom.parent() is OK                                                  # needs sage.rings.number_field
            True

        TESTS::

            sage: x = K.random_element()
            sage: x == x.numerator() / x.denominator()
            True
        """
        ring = (<RingExtensionFractionField>self._parent)._ring
        denom = self._backend.denominator()
        return ring(denom)


# Finite free extensions
########################

cdef class RingExtensionWithBasisElement(RingExtensionElement):
    r"""
    A class for elements lying in finite free extensions.

    TESTS::

        sage: K.<a> = GF(5^3).over()
        sage: L.<b> = GF(5^9).over(K)
        sage: type(b)
        <class 'sage.rings.ring_extension_element.RingExtensionWithBasisElement'>
        sage: TestSuite(b).run()
    """
    def __hash__(self):
        """
        Return a hash of this element.

        EXAMPLES:

            sage: E.<a> = GF(5^3).over()                                                # needs sage.rings.finite_rings
            sage: hash(a)                                                               # needs sage.rings.finite_rings
            5
        """
        return hash(self._backend)

    def _repr_extension(self, base, **options):
        r"""
        Return a string representation of this element written as
        a linear combination over ``base`` in the basis provided by
        the method :meth:`basis_over`.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: K.<a> = GF(5^3).over()
            sage: L.<b> = GF(5^9).over(K)
            sage: u = 1/(a+b)
            sage: u._repr_extension(base=K)
            '(2 + 2*a) + (-1 + a - a^2)*b + (2 + 3*a + 3*a^2)*b^2'
            sage: u._repr_extension(base=GF(5))
            '2 + 2*a - b + a*b - a^2*b + 2*b^2 + 3*a*b^2 + 3*a^2*b^2'
        """
        cdef RingExtensionWithBasis parent = self._parent
        coeffs = self._vector(base)
        names = parent._basis_names
        b = parent._base
        while b is not base:
            new_names = [ ]
            for y in names:
                for x in (<RingExtensionWithBasis>b)._basis_names:
                    if x == "":
                        new_names.append(y)
                    elif y == "":
                        new_names.append(x)
                    else:
                        new_names.append(x + "*" + y)
            names = new_names
            b = (<RingExtensionWithBasis>b)._base
        s = ""
        for i in range(len(names)):
            c = coeffs[i]
            if c.is_zero():
                continue
            sign = 1
            ss = ""
            if c == -1:
                sign = -1
            elif c != 1:
                atomic = c._is_atomic()
                if not atomic and (-c)._is_atomic():
                    c = -c
                    sign = -sign
                    atomic = True
                sc = str(c)
                if atomic:
                    ss += sc
                else:
                    ss += "(" + sc + ")"
                if names[i] != "":
                    ss += "*"
            if ss and ss[0] == "-":
                ss = ss[1:]
                sign *= -1
            if s == "":
                if sign == -1:
                    s = "-"
            else:
                s += " + " if sign == 1 else " - "
            ss += names[i]
            if ss == "":
                ss += "1"
            s += ss
        if s == "":
            return "0"
        if s[0] == "(" and s[-1] == ")":
            s = s[1:-1]
        return s

    def _latex_extension(self, base, **options):
        r"""
        Return a LaTeX representation of this element written as
        a linear combination over ``base`` in the basis provided by
        the method :meth:`basis_over`.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: K.<a> = GF(5^3).over()
            sage: L.<b> = GF(5^9).over(K)
            sage: u = 1/(a+b)
            sage: u._latex_extension(base=K)
            \left( 2 + 2 a \right) + \left( -1 + a - a^{2} \right) b + \left( 2 + 3 a + 3 a^{2} \right) b^{2}
            sage: u._latex_extension(base=GF(5))
            2 + 2 a - b + ab - a^{2}b + 2 b^{2} + 3 ab^{2} + 3 a^{2}b^{2}
        """
        cdef RingExtensionWithBasis parent = self._parent
        coeffs = self._vector(base)
        names = parent._basis_latex_names
        b = parent._base
        while b is not base:
            names = [ x + y for y in names for x in (<RingExtensionWithBasis>b)._basis_latex_names ]
            b = (<RingExtensionWithBasis>b)._base
        s = ""
        for i in range(len(names)):
            c = coeffs[i]
            if c.is_zero():
                continue
            sign = 1
            ss = ""
            if c == -1:
                sign = -1
            elif c != 1:
                atomic = c._is_atomic()
                if not atomic and (-c)._is_atomic():
                    c = -c
                    sign = -sign
                    atomic = True
                sc = latex(c)
                if atomic:
                    ss += sc
                else:
                    ss += r"\left(" + sc + r"\right)"
            if ss != "" and ss[0] == "-":
                ss = ss[1:]
                sign *= -1
            if s == "":
                if sign == -1:
                    s = "-"
            else:
                s += " + " if sign == 1 else " - "
            ss += names[i]
            if ss == "":
                ss += "1"
            s += ss
        if s == "":
            return "0"
        if s[:6] == r"\left(" and s[-7] == r"\right)":
            s = s[6:-7]
        return s

    def vector(self, base=None):
        r"""
        Return the vector of coordinates of this element over ``base``
        (in the basis output by the method :meth:`basis_over`).

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: F = GF(5)
            sage: K.<a> = GF(5^2).over()  # over F
            sage: L.<b> = GF(5^6).over(K)
            sage: x = (a+b)^4; x
            (-1 + a) + (3 + a)*b + (1 - a)*b^2
            sage: x.vector(K)                   # basis is (1, b, b^2)
            (-1 + a, 3 + a, 1 - a)
            sage: x.vector(F)                   # basis is (1, a, b, a*b, b^2, a*b^2)
            (4, 1, 3, 1, 1, 4)

        If ``base`` is omitted, it is set to its default which is the
        base of the extension::

            sage: x.vector()                                                            # needs sage.rings.finite_rings
            (-1 + a, 3 + a, 1 - a)

        Note that ``base`` must be an explicit base over which the
        extension has been defined (as listed by the method :meth:`bases`)::

            sage: x.vector(GF(5^3))                                                     # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            ValueError: not (explicitly) defined over Finite Field in z3 of size 5^3
        """
        base = (<RingExtension_generic>self._parent)._check_base(base)
        return self._vector(base)

    cdef _vector(self, Parent base):
        r"""
        Return the vector of coordinates of this element over ``base``
        (in the basis output by the method :meth:`basis_over`).

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        TESTS::

            sage: K = GF(11^10).over(GF(11^2))
            sage: x = K.random_element()
            sage: coeffs = x.vector()
            sage: basis = K.basis_over()
            sage: x == sum(coeffs[i]*basis[i] for i in range(5))
            True
        """
        _, _, j = (<RingExtensionWithBasis>self._parent)._free_module(base, map=True)
        return j(self)

    def polynomial(self, base=None, var='x'):
        r"""
        Return a polynomial (in one or more variables) over ``base``
        whose evaluation at the generators of the parent equals this
        element.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: F.<a> = GF(5^2).over()  # over GF(5)
            sage: K.<b> = GF(5^4).over(F)
            sage: L.<c> = GF(5^12).over(K)
            sage: u = 1/(a + b + c); u
            (2 + (-1 - a)*b) + ((2 + 3*a) + (1 - a)*b)*c + ((-1 - a) - a*b)*c^2
            sage: P = u.polynomial(K); P
            ((-1 - a) - a*b)*x^2 + ((2 + 3*a) + (1 - a)*b)*x + 2 + (-1 - a)*b
            sage: P.base_ring() is K
            True
            sage: P(c) == u
            True

        When the base is `F`, we obtain a bivariate polynomial::

            sage: P = u.polynomial(F); P                                                # needs sage.rings.finite_rings
            (-a)*x0^2*x1 + (-1 - a)*x0^2 + (1 - a)*x0*x1 + (2 + 3*a)*x0 + (-1 - a)*x1 + 2

        We check that its value at the generators is the element we started with::

            sage: L.gens(F)                                                             # needs sage.rings.finite_rings
            (c, b)
            sage: P(c, b) == u                                                          # needs sage.rings.finite_rings
            True

        Similarly, when the base is ``GF(5)``, we get a trivariate polynomial:

            sage: P = u.polynomial(GF(5)); P                                            # needs sage.rings.finite_rings
            -x0^2*x1*x2 - x0^2*x2 - x0*x1*x2 - x0^2 + x0*x1 - 2*x0*x2 - x1*x2 + 2*x0 - x1 + 2
            sage: P(c, b, a) == u                                                       # needs sage.rings.finite_rings
            True

        Different variable names can be specified::

            sage: u.polynomial(GF(5), var='y')                                          # needs sage.rings.finite_rings
            -y0^2*y1*y2 - y0^2*y2 - y0*y1*y2 - y0^2 + y0*y1 - 2*y0*y2 - y1*y2 + 2*y0 - y1 + 2
            sage: u.polynomial(GF(5), var=['x','y','z'])                                # needs sage.rings.finite_rings
            -x^2*y*z - x^2*z - x*y*z - x^2 + x*y - 2*x*z - y*z + 2*x - y + 2

        If ``base`` is omitted, it is set to its default which is the
        base of the extension::

            sage: u.polynomial()                                                        # needs sage.rings.finite_rings
            ((-1 - a) - a*b)*x^2 + ((2 + 3*a) + (1 - a)*b)*x + 2 + (-1 - a)*b

        Note that ``base`` must be an explicit base over which the
        extension has been defined (as listed by the method :meth:`bases`)::

            sage: u.polynomial(GF(5^3))                                                 # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            ValueError: not (explicitly) defined over Finite Field in z3 of size 5^3
        """
        base = self._parent._check_base(base)
        degrees = [ ]
        b = self._parent
        degree = 1
        while b is not base:
            if not isinstance(b, RingExtensionWithGen):
                raise NotImplementedError
            reldeg = b.relative_degree()
            degree *= reldeg
            degrees.append(reldeg)
            b = b.base_ring()
        degrees.reverse()
        coeffs = { }
        v = self._vector(base)
        S = PolynomialRing(base, len(degrees), names=var)
        for i in range(degree):
            ii = ZZ(i)
            exponents = [ ]
            for d in degrees:
                ii, exponent = ii.quo_rem(d)
                exponents.append(exponent)
            coeffs[tuple(reversed(exponents))] = v[i]
        return S(coeffs)

    def matrix(self, base=None):
        r"""
        Return the matrix of the multiplication by this element (in
        the basis output by :meth:`basis_over`).

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: K.<a> = GF(5^3).over()  # over GF(5)
            sage: L.<b> = GF(5^6).over(K)
            sage: u = a/(1+b)
            sage: u
            (2 + a + 3*a^2) + (3 + 3*a + a^2)*b
            sage: b*u
            (3 + 2*a^2) + (2 + 2*a - a^2)*b
            sage: u.matrix(K)
            [2 + a + 3*a^2 3 + 3*a + a^2]
            [    3 + 2*a^2 2 + 2*a - a^2]
            sage: u.matrix(GF(5))
            [2 1 3 3 3 1]
            [1 3 1 2 0 3]
            [2 3 3 1 3 0]
            [3 0 2 2 2 4]
            [4 2 0 3 0 2]
            [0 4 2 4 2 0]

        If ``base`` is omitted, it is set to its default which is the
        base of the extension::

            sage: u.matrix()                                                            # needs sage.rings.finite_rings
            [2 + a + 3*a^2 3 + 3*a + a^2]
            [    3 + 2*a^2 2 + 2*a - a^2]

        Note that ``base`` must be an explicit base over which the
        extension has been defined (as listed by the method :meth:`bases`)::

            sage: u.matrix(GF(5^2))                                                     # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            ValueError: not (explicitly) defined over Finite Field in z2 of size 5^2
        """
        cdef RingExtension_generic parent = self._parent
        base = parent._check_base(base)
        if not (parent._is_finite_over(base) and parent._is_free_over(base)):
            raise ValueError("the extension is not finite free")
        return self._matrix(base)

    cdef _matrix(self, Parent base):
        r"""
        Return the matrix of the multiplication by this element (in
        the basis output by :meth:`basis_over`).

        This method does not check its input.
        Do not call it directly; use :meth:`matrix` instead.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension)

        TESTS::

            sage: F = GF(11^2)
            sage: K = GF(11^6).over(F)
            sage: L = GF(11^18).over(K)
            sage: for base in L.bases():
            ....:     x = L.random_element()
            ....:     y = L.random_element()
            ....:     assert((x+y).matrix(base) == x.matrix(base) + y.matrix(base))
            ....:     assert((x*y).matrix(base) == x.matrix(base) * y.matrix(base))
        """
        from sage.matrix.matrix_space import MatrixSpace
        cdef RingExtensionWithBasis parent = self._parent
        _, _, j = parent._free_module(base, map=True)
        x = self._backend
        M = [ j(x * (<RingExtensionElement>b)._backend) for b in parent._basis_over(base) ]
        return MatrixSpace(base, len(M))(M)

    def trace(self, base=None):
        r"""
        Return the trace of this element over ``base``.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: F = GF(5)
            sage: K.<a> = GF(5^3).over(F)
            sage: L.<b> = GF(5^6).over(K)
            sage: u = a/(1+b)
            sage: tr = u.trace(K); tr
            -1 + 3*a + 2*a^2

        We check that the trace lives in the base ring::

            sage: tr.parent()                                                           # needs sage.rings.finite_rings
            Field in a with defining polynomial x^3 + 3*x + 3 over its base
            sage: tr.parent() is K                                                      # needs sage.rings.finite_rings
            True

        Similarly, one can compute the trace over F::

            sage: u.trace(F)                                                            # needs sage.rings.finite_rings
            0

        We check the transitivity of the trace::

            sage: u.trace(F) == tr.trace(F)                                             # needs sage.rings.finite_rings
            True

        If ``base`` is omitted, it is set to its default which is the
        base of the extension::

            sage: u.trace()                                                             # needs sage.rings.finite_rings
            -1 + 3*a + 2*a^2

        Note that ``base`` must be an explicit base over which the
        extension has been defined (as listed by the method :meth:`bases`)::

            sage: u.trace(GF(5^2))                                                      # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            ValueError: not (explicitly) defined over Finite Field in z2 of size 5^2
        """
        cdef RingExtension_generic parent = self._parent
        base = parent._check_base(base)
        if not (parent._is_finite_over(base) and parent._is_free_over(base)):
            raise ValueError("the extension is not finite free")
        return self._trace(base)

    cdef _trace(self, Parent base):
        r"""
        Return the trace of this element over ``base``.

        This method does not check its input.
        Do not call it directly; use :meth:`trace` instead.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension)

        TESTS::

            sage: F = GF(11^2)
            sage: K = GF(11^6).over(F)
            sage: L = GF(11^18).over(K)
            sage: x = L.random_element()
            sage: x.trace(F) == x.trace().trace()
            True
            sage: for base in L.bases():
            ....:     x = L.random_element()
            ....:     y = L.random_element()
            ....:     assert(x.trace(base) == x.matrix(base).trace())
            ....:     assert((x+y).trace(base) == x.trace(base) + y.trace(base))
        """
        cdef RingExtensionWithBasis parent = self._parent
        cdef Parent b
        if base is parent:
            return self
        b = parent._base
        t = self._matrix(b).trace()
        if base is b:
            return t
        return (<RingExtensionWithBasisElement>t)._trace(base)

    def norm(self, base=None):
        r"""
        Return the norm of this element over ``base``.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: F = GF(5)
            sage: K.<a> = GF(5^3).over(F)
            sage: L.<b> = GF(5^6).over(K)
            sage: u = a/(1+b)
            sage: nr = u.norm(K); nr
            3 + 2*a^2

        We check that the norm lives in the base ring::

            sage: nr.parent()                                                           # needs sage.rings.finite_rings
            Field in a with defining polynomial x^3 + 3*x + 3 over its base
            sage: nr.parent() is K                                                      # needs sage.rings.finite_rings
            True

        Similarly, one can compute the norm over F::

            sage: u.norm(F)                                                             # needs sage.rings.finite_rings
            4

        We check the transitivity of the norm::

            sage: u.norm(F) == nr.norm(F)                                               # needs sage.rings.finite_rings
            True

        If ``base`` is omitted, it is set to its default which is the
        base of the extension::

            sage: u.norm()                                                              # needs sage.rings.finite_rings
            3 + 2*a^2

        Note that ``base`` must be an explicit base over which the
        extension has been defined (as listed by the method :meth:`bases`)::

            sage: u.norm(GF(5^2))                                                       # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            ValueError: not (explicitly) defined over Finite Field in z2 of size 5^2
        """
        cdef RingExtension_generic parent = self._parent
        base = parent._check_base(base)
        if not (parent._is_finite_over(base) and parent._is_free_over(base)):
            raise ValueError("the extension is not finite free")
        return self._norm(base)

    cdef _norm(self, Parent base):
        r"""
        Return the norm of this element over ``base``.

        This method does not check its input.
        Do not call it directly; use :meth:`norm` instead.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension)

        TESTS::

            sage: F = GF(11^2)
            sage: K = GF(11^6).over(F)
            sage: L = GF(11^18).over(K)
            sage: x = L.random_element()
            sage: x.norm(F) == x.norm().norm()
            True
            sage: for base in L.bases():
            ....:     x = L.random_element()
            ....:     y = L.random_element()
            ....:     assert(x.norm(base) == x.matrix(base).determinant())
            ....:     assert((x*y).norm(base) == x.norm(base) * y.norm(base))
        """
        cdef RingExtensionWithBasis parent = self._parent
        cdef Parent b
        if base is parent:
            return self
        b = parent._base
        n = self._matrix(b).determinant()
        if base is b:
            return n
        return (<RingExtensionWithBasisElement>n)._norm(base)

    def charpoly(self, base=None, var='x'):
        r"""
        Return the characteristic polynomial of this element over ``base``.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: F = GF(5)
            sage: K.<a> = GF(5^3).over(F)
            sage: L.<b> = GF(5^6).over(K)
            sage: u = a/(1+b)
            sage: chi = u.charpoly(K); chi
            x^2 + (1 + 2*a + 3*a^2)*x + 3 + 2*a^2

        We check that the charpoly has coefficients in the base ring::

            sage: chi.base_ring()                                                       # needs sage.rings.finite_rings
            Field in a with defining polynomial x^3 + 3*x + 3 over its base
            sage: chi.base_ring() is K                                                  # needs sage.rings.finite_rings
            True

        and that it annihilates u::

            sage: chi(u)                                                                # needs sage.rings.finite_rings
            0

        Similarly, one can compute the characteristic polynomial over F::

            sage: u.charpoly(F)                                                         # needs sage.rings.finite_rings
            x^6 + x^4 + 2*x^3 + 3*x + 4

        A different variable name can be specified::

            sage: u.charpoly(F, var='t')                                                # needs sage.rings.finite_rings
            t^6 + t^4 + 2*t^3 + 3*t + 4

        If ``base`` is omitted, it is set to its default which is the
        base of the extension::

            sage: u.charpoly()                                                          # needs sage.rings.finite_rings
            x^2 + (1 + 2*a + 3*a^2)*x + 3 + 2*a^2

        Note that ``base`` must be an explicit base over which the
        extension has been defined (as listed by the method :meth:`bases`)::

            sage: u.charpoly(GF(5^2))                                                   # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            ValueError: not (explicitly) defined over Finite Field in z2 of size 5^2

        TESTS:

        We check that the characteristic polynomial of an element in the base
        ring is a power of a polynomial of degree 1::

            sage: S.<x> = K[]                                                           # needs sage.rings.finite_rings
            sage: u = K.random_element()                                                # needs sage.rings.finite_rings
            sage: L(u).charpoly() == (x - u)^2                                          # needs sage.rings.finite_rings
            True
        """
        return self.matrix(base).charpoly(var)

    cpdef minpoly(self, base=None, var='x'):
        r"""
        Return the minimal polynomial of this element over ``base``.

        INPUT:

        - ``base`` -- a commutative ring (which might be itself an
          extension) or ``None``

        EXAMPLES::

            sage: F = GF(5)
            sage: K.<a> = GF(5^3).over(F)
            sage: L.<b> = GF(5^6).over(K)
            sage: u = 1 / (a+b)
            sage: chi = u.minpoly(K); chi
            x^2 + (2*a + a^2)*x - 1 + a

        We check that the minimal polynomial has coefficients in the base ring::

            sage: chi.base_ring()                                                       # needs sage.rings.finite_rings
            Field in a with defining polynomial x^3 + 3*x + 3 over its base
            sage: chi.base_ring() is K                                                  # needs sage.rings.finite_rings
            True

        and that it annihilates u::

            sage: chi(u)                                                                # needs sage.rings.finite_rings
            0

        Similarly, one can compute the minimal polynomial over F::

            sage: u.minpoly(F)                                                          # needs sage.rings.finite_rings
            x^6 + 4*x^5 + x^4 + 2*x^2 + 3

        A different variable name can be specified::

            sage: u.minpoly(F, var='t')                                                 # needs sage.rings.finite_rings
            t^6 + 4*t^5 + t^4 + 2*t^2 + 3

        If ``base`` is omitted, it is set to its default which is the
        base of the extension::

            sage: u.minpoly()                                                           # needs sage.rings.finite_rings
            x^2 + (2*a + a^2)*x - 1 + a

        Note that ``base`` must be an explicit base over which the
        extension has been defined (as listed by the method :meth:`bases`)::

            sage: u.minpoly(GF(5^2))                                                    # needs sage.rings.finite_rings
            Traceback (most recent call last):
            ...
            ValueError: not (explicitly) defined over Finite Field in z2 of size 5^2

        TESTS:

        We check that the minimal polynomial of an element in the base
        ring has degree 1::

            sage: S.<x> = K[]                                                           # needs sage.rings.finite_rings
            sage: u = K.random_element()                                                # needs sage.rings.finite_rings
            sage: L(u).minpoly() == x - u                                               # needs sage.rings.finite_rings
            True

        In a similar fashion, the minimal polynomial over `F` of an element
        of `K` should have degree 1 or 3::

            sage: L(u).minpoly(F).degree() in [ 1, 3 ]                                  # needs sage.rings.finite_rings
            True
        """
        cdef RingExtensionWithBasis parent = self._parent

        if base is None:
            mod = parent.modulus()
            S = mod.parent().quotient(mod)
            try:
                return S(list(self.vector())).minpoly()
            except NotImplementedError:
                pass  # fall back to generic code below

        from sage.modules.free_module import FreeModule
        cdef MapRelativeRingToFreeModule j

        base = parent._check_base(base)
        if not (parent._is_finite_over(base) and parent._is_free_over(base)):
            raise ValueError("the extension is not finite free")
        if base not in Fields():
            raise NotImplementedError("minpoly is only implemented when the base is a field")
        K = backend_parent(base)
        degree = parent._degree_over(base)
        _, _, j = parent._free_module(base, map=True)
        V = FreeModule(K, degree)
        vector = [K(1)] + (degree-1)*[K(0)]
        vectors = [vector]
        W = V.span(vectors)
        elt = self
        while True:
            vector = V(j.backend_coefficients(elt))
            if vector in W: break
            vectors.append(vector)
            W += V.span([vector])
            elt *= self
        W = V.span_of_basis(vectors)
        coeffs = [ -c for c in W.coordinate_vector(vector) ] + [K(1)]
        return PolynomialRing(base, name=var)(coeffs)
